# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from openexp._canvas._element.group import Group
from openexp.canvas_elements import Line, Ellipse


class FixDot(Group):

	def __init__(self, canvas, x=None, y=None, style=u'default', **properties):

		if x is None:
			x = 0 if canvas.uniform_coordinates else canvas._width/2
		if y is None:
			y = 0 if canvas.uniform_coordinates else canvas._height/2
		h = 2
		if u'large' in style:
			s = 16
		elif u'medium' in style or style == u'default':
			s = 8
		elif u'small' in style:
			s = 4
		else:
			raise osexception(u'Unknown style: %s' % self.style)
		if u'open' in style or style == u'default':
			elements = [
				Ellipse(x-s, y-s, 2*s, 2*s, fill=True, **properties) \
					.construct(canvas),
				Ellipse(x-h, y-h, 2*h, 2*h, fill=True,
					color=canvas.background_color.colorspec,
					**{key : val for key, val in properties.items() \
					if key != u'color'}).construct(canvas)
				]
		elif u'filled' in style:
			elements = [Ellipse(
				x-s, y-s, 2*s, 2*s, fill=True, **properties)\
				.construct(canvas)]
		elif u'cross' in style:
			elements = [
				Line(x, y-s, x, y+s, **properties).construct(canvas),
				Line(x-s, y, x+s, y, **properties).construct(canvas)
				]
		else:
			raise osexception(u'Unknown style: %s' % self.style)
		properties = properties.copy()
		properties.update({ 'x' : x, 'y' : y, u'style' : style })
		Group.__init__(self, canvas, elements, **properties)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import platform
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'system_information', category=u'extension')


class system_information(base_extension):

	"""
	desc:
		Shows a tab with system information and version numbers
	"""

	def activate(self):

		l = []
		for name, module, attr in [
			(u'OpenSesame', u'libopensesame.metadata', u'__version__'),
			(u'Python', u'sys', u'version'),
			(u'datamatrix', u'datamatrix', u'__version__'),
			(u'qdatamatrix', u'qdatamatrix', u'__version__'),
			(u'pseudorandom', u'pseudorandom', u'__version__'),
			(u'fileinspector', u'fileinspector', u'__version__'),
			(u'QNotifications', u'QNotifications', u'__version__'),
			(u'QOpenScienceFramework', u'QOpenScienceFramework', u'__version__'),
			(u'opencv', u'cv2', u'__version__'),
			(u'expyriment', u'expyriment', u'__version__'),
			(u'IPython', u'IPython', u'__version__'),
			(u'numpy', u'numpy', u'__version__'),
			(u'scipy', u'scipy', u'__version__'),
			(u'PIL/ PILLOW', u'PIL', u'VERSION'),
			(u'psychopy', u'psychopy', u'__version__'),
			(u'pygame', u'pygame', u'__version__'),
			(u'pygaze', u'pygaze', u'__version__'),
			(u'pyglet', u'pyglet', u'version'),
			(u'PyQt', u'qtpy', [u'QtCore', u'PYQT_VERSION_STR']),
			(u'serial', u'serial', u'VERSION'),
			(u'markdown', u'markdown', u'version'),
			(u'yaml', u'yaml', u'__version__'),
			]:
			try:
				mod = __import__(module)
			except ImportError:
				l.append(u'%s [not available]' % name)
				continue
			except Exception as e:
				l.append(
					u'%s [installed, but `%s` occured while importing]' \
					% (name, e.__class__.__name__))
				continue
			if isinstance(attr, basestring):
				attr = [attr]
			while len(attr) > 1:
				a = attr.pop(0)
				if hasattr(mod, a):
					mod = getattr(mod, a)
			a = attr.pop(0)
			if hasattr(mod, a):
				l.append(u'%s %s' % (name, getattr(mod, a)))
				continue
			l.append(u'%s [version unknown]' % name)

		# platform.architecture() does not seem to work correctly on Windows so a solution found on stackoverflow is used
		# http://stackoverflow.com/questions/2208828/detect-64bit-os-windows-in-python
		if(platform.platform().startswith('Windows') and platform.machine().endswith('64')):
			t = 'win64'
		elif(platform.platform().startswith('Windows') and not platform.machine().endswith('64')):
			t = 'win32'
		else:
			t = platform.architecture()[0]

		md = safe_read(self.ext_resource(u'system-information.md')) % {
			u'system' : platform.platform(),
			u'architecture' : t,
			u'modules' : u'- ' + u'\n- '.join(l)
			}
		self.tabwidget.open_markdown(md, icon=self.icon(),
			title=_(u'System information'))



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtGui
from libqtopensesame.misc.base_subcomponent import base_subcomponent

class base_validator(base_subcomponent, QtGui.QValidator):

	"""
	desc:
		A base class for input validators.
	"""

	def __init__(self, main_window, default):

		self.default = default
		main_window = self.get_main_window(main_window)
		super(base_validator, self).__init__(main_window)
		self.setup(main_window)

	def is_valid(self, val):

		return True

	def validate(self, val, pos):

		if self.is_valid(val):
			return self.Acceptable, val, pos
		return self.Intermediate, val, pos

	def fixup(self, val):

		if not self.is_valid(val):
			return self.default
		return val



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._sampler.sampler import Sampler
from libopensesame.exceptions import osexception
from libopensesame.oslogging import oslogger
from openexp.backend import configurable
from openexp.keyboard import Keyboard
import soundfile as sf
import numpy as np

DEFAULT_SOUND_FREQ = 48000
DEFAULT_BLOCK_SIZE = 256
NEEDS_BLOCK_SIZE = 'sounddevice', 'PTB'
# Will be intialized during init_sound()
Sound = None
PLAYING = None


class Psycho(Sampler):

	"""
	desc:
		This is a sampler backend built on top of PsychoPy. For function
		specifications and docstrings, see `openexp._sampler.sampler`.
	"""

	# The settings variable is used by the GUI to provide a list of back-end
	# settings
	settings = {
		'psycho_audiolib': {
			'name': 'Sound library',
			'description': 'Can be sounddevice, pyo, pygame, or PTB',
			'default': 'sounddevice'
		},
		'sound_freq': {
			'name': 'Sampling frequency for synth',
			'description': 'Determines the sampling rate of synthesized sounds',
			'default': DEFAULT_SOUND_FREQ
		},
		'block_size': {
			'name': 'Buffer size',
			'description': 'Low values for low latency. High values for stability.',
			'default': DEFAULT_BLOCK_SIZE
		}
	}

	def __init__(self, experiment, src, **playback_args):

		if (
			isinstance(src, np.ndarray) and
			src.dtype == np.int16 and
			src.ndim == 1
		):
			# The Synth provides the data as a 1D array with int values between
			# 0 and 32767. The signal is stereo but flattened into a single
			# trace, so we expand it here.
			self._data = np.array(src, dtype=np.float64) / 32767
			self._data = self._data.reshape((self._data.shape[0] // 2, 2))
			self._samplerate = experiment.var.get(
				'sound_freq',
				DEFAULT_SOUND_FREQ
			)
		else:
			self._data, self._samplerate = sf.read(src)
		# Create keyword arguments, which depend on the sound backend
		kwargs = {}
		if (
			experiment.var.get('psycho_audiolib', 'sounddevice')
			in NEEDS_BLOCK_SIZE
		):			
			kwargs['blockSize'] = experiment.var.get(
				'block_size',
				DEFAULT_BLOCK_SIZE
			)
		# Make sure that the data is a [samples, 2] array for stereo data
		if self._data.ndim == 1:
			self._data.shape = [len(self._data), 1]
		if self._data.shape[1] == 1:
			self._data = self._data.repeat(2, axis=1)
		self._sound = Sound(
			self._data,
			sampleRate=self._samplerate,
			**kwargs
		)
		self._keyboard = Keyboard(experiment)
		Sampler.__init__(self, experiment, src, **playback_args)

	def set_config(self, **cfg):

		Sampler.set_config(self, **cfg)
		if not cfg:
			return
		if 'volume' in cfg:
			self._set_volume(cfg['volume'])
		if 'pitch' in cfg:
			self._set_pitch(cfg['pitch'])
		if 'pan' in cfg:
			self._set_pan(cfg['pan'])
		if 'duration' in cfg:
			self._set_duration(cfg['duration'])
		if 'fade_in' in cfg:
			self._set_fade_in(cfg['fade_in'])
		self._sound.stopTime = -1  # Fixes a bug in PsychoPy 3.2.4
		self._sound.setSound(self._data)

	def _set_volume(self, volume):

		if volume == 1:
			return
		self._data *= volume

	def _set_pitch(self, pitch):

		from scipy.signal import resample
		if pitch == 1:
			return
		self._data = resample(self._data, int(self._data.shape[0] / pitch))

	def _set_pan(self, pan):

		if pan == 'left':
			right_volume = 0.
			left_volume = 1.
		elif pan == 'right':
			right_volume = 1.
			left_volume = 0.
		elif pan < 0:
			right_volume = -1. / pan
			left_volume = 1.
		elif pan > 0:
			right_volume = 1.
			left_volume = 1. / pan
		else:
			return
		self._data[:, 0] *= left_volume
		self._data[:, 1] *= right_volume

	def _ms_to_samples(self, ms):

		return int(self._samplerate * (ms / 1000))

	def _set_duration(self, duration):

		if not duration:
			return
		self._data = self._data[:self._ms_to_samples(duration), :]

	def _set_fade_in(self, fade_in):

		if fade_in is None:
			return
		ramp = np.linspace(
			0, 1, self._ms_to_samples(fade_in)
		)[:self._data.shape[0]]
		self._data[:len(ramp), 0] *= ramp
		self._data[:len(ramp), 1] *= ramp

	@configurable
	def play(self, **playback_args):

		self._sound.stop()
		self._sound.play()
		if self.block:
			self.wait()

	def stop(self):

		self._sound.stop()

	def pause(self):

		self._sound.pause()

	def resume(self):

		self._sound.play()

	def is_playing(self):

		return self._sound.status == PLAYING

	def wait(self):

		while self._sound.status == PLAYING:
			self._keyboard.flush()

	@staticmethod
	def init_sound(experiment):

		global Sound, PLAYING

		from psychopy import prefs
		prefs.hardware['audioLib'] = [
			experiment.var.get('psycho_audiolib', 'sounddevice')
		]
		from psychopy import constants
		PLAYING = constants.PLAYING
		from psychopy.sound import Sound

	@staticmethod
	def close_sound(experiment):

		pass


# Non PEP-8 alias for backwards compatibility
psycho = Psycho



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from pyqode_extras.modes.convert_indentation_mode import ConvertIndentationMode
from pyqode_extras.modes.autodetect_indentation_mode import AutodetectIndentationMode



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy.QtWidgets import QSizePolicy
from qtpy.QtCore import Qt
from libopensesame.inline_script import inline_script as inline_script_runtime
from libqtopensesame.items.qtplugin import qtplugin
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'inline_script', category=u'item')


class inline_script(inline_script_runtime, qtplugin):

	"""The inline_script GUI controls"""

	description = _(u'Executes Python code')
	help_url = u'manual/python/about'
	ext = u'.py'
	mime_type = u'text/x-python'

	def __init__(self, name, experiment, string=None):

		"""See item."""

		inline_script_runtime.__init__(self, name, experiment, string)
		qtplugin.__init__(self)

	def apply_edit_changes(self):

		"""See qtitem."""

		sp = self._pyqode_prepare_editor.toPlainText()
		sr = self._pyqode_run_editor.toPlainText()
		self._set_modified()
		self.var._prepare = sp
		self.var._run = sr
		qtplugin.apply_edit_changes(self)

	def _set_modified(self, prepare=False, run=False):

		"""
		desc:
			Sets the modified status of the editors.
		"""

		self._pyqode_prepare_editor.document().setModified(prepare)
		self._pyqode_run_editor.document().setModified(run)
		self._pyqode_tab_bar.setTabText(
			0,
			(u'* ' if prepare else u'') + _(u'Prepare')
		)
		self._pyqode_tab_bar.setTabText(
			1,
			(u'* ' if prepare else u'') + _(u'Run')
		)

	def set_focus(self):

		"""
		desc:
			Allows the item to focus the most important widget.
		"""

		self._pyqode_tab_widget.setFocus()

	def init_edit_widget(self):

		"""See qtitem."""

		from pyqode.core.widgets import SplittableCodeEditTabWidget

		qtplugin.init_edit_widget(self, stretch=False)
		self._pyqode_tab_widget = SplittableCodeEditTabWidget()
		self._pyqode_tab_widget.setSizePolicy(
			QSizePolicy.Expanding,
			QSizePolicy.Expanding
		)
		self._pyqode_tab_widget.main_tab_widget.setCornerWidget(None)
		self._pyqode_tab_bar = self._pyqode_tab_widget.main_tab_widget.tabBar()
		self._pyqode_tab_bar.setTabsClosable(False)
		self._pyqode_tab_bar.setContextMenuPolicy(Qt.NoContextMenu)
		self._pyqode_tab_widget.main_tab_widget.setMovable(False)
		self._pyqode_prepare_editor = \
			self._pyqode_tab_widget.create_new_document('Prepare', self.ext)
		self._pyqode_run_editor = \
			self._pyqode_tab_widget.create_new_document('Run', self.ext)
		self._pyqode_run_editor.focusOutEvent = self._editor_focus_out
		self._pyqode_prepare_editor.focusOutEvent = self._editor_focus_out
		self._set_modified()
		self.extension_manager.fire(
			u'register_editor',
			editor=self._pyqode_run_editor
		)
		self.extension_manager.fire(
			u'register_editor',
			editor=self._pyqode_prepare_editor
		)
		self.edit_vbox.addWidget(self._pyqode_tab_widget)
		if not self.var._run and self.var._prepare:
			self._pyqode_tab_widget.main_tab_widget.setCurrentIndex(0)
		else:
			self._pyqode_tab_widget.main_tab_widget.setCurrentIndex(1)


	def edit_widget(self):

		"""See qtitem."""

		qtplugin.edit_widget(self)
		_prepare = safe_decode(self.var._prepare)
		if _prepare != self._pyqode_prepare_editor.toPlainText():
			self._pyqode_prepare_editor.setPlainText(
				_prepare,
				mime_type=self.mime_type
			)
		_run = safe_decode(self.var._run)
		if _run != self._pyqode_run_editor.toPlainText():
			self._pyqode_run_editor.setPlainText(
				_run,
				mime_type=self.mime_type
			)

	def get_ready(self):

		if self.container_widget is None:
			return
		self.apply_edit_changes()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *


class qtstructure_item(object):

	"""
	desc:
		A base class for items that control the structure of the experiment,
		i.e. those that require rebuilding of the item tree when they are
		changed.
	"""
	
	def __init__(self):
		
		"""
		desc:
			Constructor.
		"""

		self._children = None

	def update(self):

		"""See qtitem."""

		super(qtstructure_item, self).update()
		self.experiment.build_item_tree()

	def apply_script_changes(self):

		"""See qtitem."""

		super(qtstructure_item, self).apply_script_changes()
		self.experiment.build_item_tree()
		
	@staticmethod
	def clears_children_cache(fnc):
		
		"""
		desc:
			A decorator for functions that change the structure of the
			experiment, and thus need a clearing of the children cache.
		"""
		
		def inner(self, *args, **kwargs):
			
			self.experiment.items.clear_cache()
			return fnc(self, *args, **kwargs)
			
		return inner
		
	@staticmethod
	def cached_children(fnc):
		
		"""
		desc:
			A decorator for the children function, which is cached to speed up
			performance.
		"""
		
		def inner(self, *args, **kwargs):
			
			if self._children is not None:
				return self._children
			return fnc(self, *args, **kwargs)
			
		return inner



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import debug
from libqtopensesame.misc import config
from libqtopensesame.widgets.base_widget import base_widget
from qtpy import QtWidgets, QtCore
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'unused_widget', category=u'core')

class unused_widget(base_widget):

	"""
	desc:
		The unused items widget.
	"""

	tab_name = u'__unused__'

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window:
				desc:	The main-window object.
				type:	qtopensesame
		"""

		super(unused_widget, self).__init__(main_window)
		header_hbox = QtWidgets.QHBoxLayout()
		header_hbox.addWidget(self.theme.qlabel(u"unused"))
		header_label = QtWidgets.QLabel()
		header_label.setText(_(u"<b><font size='5'>Unused</font></b>"))
		header_hbox.addWidget(header_label)
		header_hbox.addStretch()
		self.purge_button = QtWidgets.QPushButton(self.theme.qicon(u"purge"),
			_(u"Permanently delete unused items"))
		self.purge_button.setIconSize(QtCore.QSize(16, 16))
		self.purge_button.clicked.connect(self.purge_unused)
		purge_hbox = QtWidgets.QHBoxLayout()
		purge_hbox.addWidget(self.purge_button)
		purge_hbox.addStretch()
		purge_widget = QtWidgets.QWidget()
		purge_widget.setLayout(purge_hbox)
		vbox = QtWidgets.QVBoxLayout()
		vbox.addWidget(purge_widget)
		vbox.addStretch()
		self.setLayout(vbox)
		self.__unused_tab__ = True

	def purge_unused(self):

		"""
		desc:
			Purges all unused items.
		"""

		resp = QtWidgets.QMessageBox.question(self.main_window.ui.centralwidget,
			_(u"Permanently delete items?"),
			_(u"Are you sure you want to permanently delete all unused items? This action cannot be undone."),
			QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
		if resp == QtWidgets.QMessageBox.No:
			return
		self.extension_manager.fire(u'prepare_purge_unused_items')
		self.overview_area.locked = True
		for item_name in self.experiment.items.unused():
			self.experiment.items[item_name].close_tab()
			del self.experiment.items[item_name]
		self.overview_area.locked = False
		self.experiment.build_item_tree()
		self.purge_button.setDisabled(True)
		self.extension_manager.fire(u'purge_unused_items')

	def on_activate(self):

		"""
		desc:
			Is called when the widget becomes visible.
		"""

		self.purge_button.setDisabled(len(self.experiment.items.unused()) == 0)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._canvas._element.element import Element


class Polygon(Element):

	def __init__(self, canvas, vertices, **properties):

		properties = properties.copy()
		properties.update({ 'vertices' : vertices })
		self._shapely_polygon = None
		Element.__init__(self, canvas, **properties)

	@property
	def rect(self):

		left = min(x for x,y in self.vertices)
		right = max(x for x,y in self.vertices)
		top = min(y for x,y in self.vertices)
		bottom = max(y for x,y in self.vertices)
		return left, top, right-left, bottom-top

	def __contains__(self, xy):

		# Shapely is used to determine whether a point falls exactly within the
		# polygon. If shapely isn't available, we fall back to a simple bounding
		# box. The shapely polygon is stored for performance.
		try:
			from shapely.geometry import Polygon as ShapelyPolygon, Point
		except ImportError:
			return Element.__contains__(self, xy)
		if self._shapely_polygon is None:
			self._shapely_polygon = ShapelyPolygon(self.vertices)
		return self._shapely_polygon.contains(Point(*xy))

	def _on_attribute_change(self, **kwargs):

		# When a change occurs the shapely polygon (if any) needs to be
		# redetermined in __contains__
		self._shapely_polygon = None
		Element._on_attribute_change(self, **kwargs)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from libopensesame.sketchpad_elements._base_element import base_element

class circle(base_element):

	"""
	desc:
		An circle element for the sketchpad.
	"""

	def __init__(self, sketchpad, string):

		"""
		desc:
			Constructor.

		arguments:
			sketchpad:		A sketchpad object.
			string:			A definition string.
		"""

		defaults = [
			(u'x'		, None),
			(u'y'		, None),
			(u'r'		, None),
			(u'fill'	, 0),
			(u'color'	, sketchpad.var.get(u'foreground')),
			(u'penwidth', 1),
			]
		super(circle, self).__init__(sketchpad, string, defaults=defaults)

	def draw(self):

		"""
		desc:
			Draws the element to the canvas of the sketchpad.
		"""

		properties = self.eval_properties()
		return self.canvas.circle(properties[u'x'], properties[u'y'],
			properties[u'r'], fill=properties[u'fill'],
			color=properties[u'color'],
			penwidth=properties[u'penwidth'])



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import openexp.mouse
from libopensesame.mouse_response import mouse_response
from libqtopensesame.items.qtautoplugin import qtautoplugin


class touch_response(mouse_response):

	"""A simple fixation-dot plug-in."""

	description = \
		u'A grid-based response item, convenient for touch screens'

	def reset(self):

		self.var._ncol = 2
		self.var._nrow = 1
		mouse_response.reset(self)

	def process_response(self, response_args):

		"""See base_response_item."""

		response, pos, t1 = response_args
		if pos is not None:
			x, y = pos
			if self.experiment.var.uniform_coordinates == u'yes':
				_x = x+self.experiment.var.width/2
				_y = y+self.experiment.var.height/2
			else:
				_x, _y = x, y
			col = _x // (self.experiment.var.width / self.var._ncol)
			row = _y // (self.experiment.var.height / self.var._nrow)
			response = row * self.var._ncol + col + 1
		mouse_response.process_response(self, (response, pos, t1) )


class qttouch_response(touch_response, qtautoplugin):

	"""The GUI part of the plug-in. Controls are defined in info.json."""

	def __init__(self, name, experiment, script=None):

		"""
		Constructor.

		Arguments:
		name		--	The name of the plug-in.
		experiment	--	The experiment object.

		Keyword arguments:
		script		--	A definition script. (default=None)
		"""

		touch_response.__init__(self, name, experiment, script)
		qtautoplugin.__init__(self, __file__)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import warnings
from libqtopensesame.extensions import base_extension

class psychopy_monitor_center(base_extension):

	"""
	desc:
		Launches the PsychoPy monitor center.
	"""

	def activate(self):

		"""
		desc:
			Is called when the extension is activated through the menu/ toolbar
			action.
		"""

		from psychopy.monitors.MonitorCenter import MonitorCenter
		with warnings.catch_warnings(record=True) as w:
			app = MonitorCenter(0)
		app.MainLoop()



#-*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.dialogs.gabor_settings import gabor_settings
from libqtopensesame.sketchpad_elements._base_element import base_element
from libopensesame.sketchpad_elements import gabor as gabor_runtime

class gabor(base_element, gabor_runtime):

	"""
	desc:
		A gabor element.

		See base_element for docstrings and function descriptions.
	"""

	def show_edit_dialog(self):

		"""
		desc:
			The show-edit dialog for the gabor shows the settings dialog.
		"""

		d = gabor_settings(self.sketchpad._edit_widget)
		d.set_properties(self.properties)
		properties = d.get_properties()
		if properties is None:
			return
		self.properties.update(properties)
		self.sketchpad.draw()

	@classmethod
	def mouse_press(cls, sketchpad, pos):

		d = gabor_settings(sketchpad._edit_widget)
		properties = d.get_properties()
		if properties is None:
			return
		properties.update({
				u'x':			pos[0],
				u'y':			pos[1],
				u'show_if' : 	sketchpad.current_show_if()
			})
		return gabor(sketchpad, properties=properties)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import base_extension


class example(base_extension):

	"""
	desc:
		An example extension that lists all available events.
	"""

	def activate(self):

		"""
		desc:
			Is called when the extension is activated through the menu/ toolbar
			action.
		"""

		oslogger.debug(u'Example extension activated')

	# Below is a list of event handlers, which you can implement to have your
	# extension react to specific events.

	def event_startup(self):
		oslogger.debug(u'Event fired: startup')

	def event_prepare_regenerate(self):
		oslogger.debug(u'Event fired: prepare_regenerate')

	def event_regenerate(self):
		oslogger.debug(u'Event fired: regenerate')

	def event_run_experiment(self, fullscreen):
		oslogger.debug(u'Event fired: run_experiment(fullscreen=%s)' % fullscreen)

	def event_end_experiment(self, ret_val):
		oslogger.debug(u'Event fired: end_experiment')

	def event_save_experiment(self, path):
		oslogger.debug(u'Event fired: save_experiment(path=%s)' % path)

	def event_open_experiment(self, path):
		oslogger.debug(u'Event fired: open_experiment(path=%s)' % path)

	def event_close(self):
		oslogger.debug(u'Event fired: close')

	def event_prepare_open_item(self, name):
		oslogger.debug(u'Event fired: prepare_open_item(name=%s)' % name)

	def event_open_item(self, name):
		oslogger.debug(u'Event fired: open_item(name=%s)' % name)

	def event_prepare_rename_item(self, from_name, to_name):
		oslogger.debug(
			u'Event fired: prepare_rename_item(from_name=%s, to_name=%s)' \
			% (from_name, to_name))

	def event_rename_item(self, from_name, to_name):
		oslogger.debug(u'Event fired: rename_item(from_name=%s, to_name=%s)' \
			% (from_name, to_name))

	def event_new_item(self, name, _type):
		oslogger.debug(u'Event fired: new_item(name=%s, _type=%s)' % (name, _type))

	def event_change_item(self, name):
		oslogger.debug(u'Event fired: change_item(name=%s)' % name)

	def event_prepare_change_experiment(self):
		oslogger.debug(u'Event fired: prepare_change_experiment')

	def event_change_experiment(self):
		oslogger.debug(u'Event fired: change_experiment')

	def event_prepare_delete_item(self, name):
		oslogger.debug(u'Event fired: prepare_delete_item(name=%s)' % name)

	def event_delete_item(self, name):
		oslogger.debug(u'Event fired: delete_item(name=%s)' % name)

	def event_prepare_purge_unused_items(self):
		oslogger.debug(u'Event fired: prepare_purge_unused_items')

	def event_purge_unused_items(self):
		oslogger.debug(u'Event fired: purge_unused_items')

	def event_process_data_files(self, data_files):
		oslogger.debug(u'Event fired: process_data_files')



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import pygame
from openexp._canvas import canvas
from openexp._canvas._gabor.gabor import Gabor
from openexp._canvas._element.xpyriment import XpyrimentElement
from expyriment.stimuli._visual import Visual


class Xpyriment(XpyrimentElement, Gabor):

	def prepare(self):

		im = canvas._gabor(
			self.orient,
			self.freq,
			self.env,
			self.size,
			self.stdev,
			self.phase,
			self.col1,
			self.col2,
			self.bgmode
		)
		surface = pygame.image.fromstring(im.tobytes(), im.size, u'RGB')
		self._stim = Visual(position=self.to_xy(self.x, self.y))
		self._stim.set_surface(surface)
		self._stim.preload()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

import shlex
import re
import codecs
import os
import yaml
from distutils.version import StrictVersion
from libopensesame import metadata
from libopensesame.exceptions import osexception
from libopensesame.py3compat import *

class syntax(object):

	"""
	desc:
		The `syntax` class implement text operations, including those that are
		necessary for interpreting OpenSesame script.
	"""

	def __init__(self, experiment):

		"""
		desc:
			Constructor.

		arguments:
			experiment:	The experiment object.
		"""

		self.experiment = experiment
		# A regular expression to match keywords, which are characterized by:
		# - Start with a letter or underscore
		# - Consists only of letters, numbers, or underscores, except for
		# - The last character, which is an equals sign
		self.re_cmd = re.compile(r'\A[_a-zA-Z]+[_a-zA-Z0-9]*=')
		# A regular expression to match [variables] in text. The first part
		# indictates that we don't match items preceded by a slash
		# (\[variables]). This introduces a problem in that the slash itself can
		# also be escaped. To deal with this we eat up any even number of
		# slashes (\\\\)* so that only sequences preceded by an odd number of
		# slashes are preserved.
		self.re_txt = re.compile(r'(?<!\\)(\\\\)*(\[[_a-zA-Z]+[_a-zA-Z0-9]*\])')
		# A regular expression to match inline Python statements, like so:
		# [=10*10]
		# [\[test\]]
		self.re_txt_py = re.compile(r'(?<!\\)(\\\\)*(\[=.*?[^\\]\])')
		# Catch single equals signs
		self.re_single_eq = re.compile(r'(?<![=!<>])(=)(?!=)')
		# Catch 'never' and 'always'
		self.re_never = re.compile(r'\bnever\b', re.I)
		self.re_always = re.compile(r'\balways\b', re.I)
		# Strict no-variables is used to remove all characters except
		# alphanumeric ones
		self.re_sanitize_strict_novars = re.compile(r'[^\w]')
		# Strict with variables is used to remove all characters except
		# alphanumeric ones and [] signs that indicate variables
		self.re_sanitize_strict_vars = re.compile(r'[^\w\[\]]')
		# Matches valid variable names
		self.re_valid_var_name = re.compile(r'\A[_a-zA-Z]+[_a-zA-Z0-9]*\Z')
		# Loose is used to remove double-quotes, slashes, and newlines
		self.re_sanitize_loose = re.compile(r'[\n\"\\]')
		# Unsanitization is used to replace U+XXXX unicode notation
		self.re_from_ascii = re.compile(r'U\+([A-F0-9]{4})')
		self.re_front_matter = re.compile(r'---(?P<info>.*?)---', re.S)

	def auto_type(self, val):

		"""
		desc:
			Casts a value to its best-fitting type, i.e. float, int, or unicode.

		arguments:
			val:	A value of any type.

		returns:
			An auto-typed value.
		"""

		try:
			f = float(val)
			# No flanking whitespace
			if isinstance(val, basestring):
				assert(val == val.strip())
		except:
			return safe_decode(val, errors=u'ignore')
		if int(f) == f:
			return int(f)
		return f

	def parse_front_matter(self, s):

		"""
		desc:
			Parses the YAML front matter at the start of the script.

		arguments:
			s:
				desc:	The script.
				type:	str

		returns:
			desc:	A (d, s) tuple, where d is a dict with the front matter, and
					and s is the script with the front matter stripped off.
			type:	dict
		"""

		m = self.re_front_matter.match(s)
		try:
			d = safe_yaml_load(m.group(u'info'))
			s = s[len(m.group(0)):]
		except:
			d = {}
		# The API can be stored in a few ways:
		# - Not at all, which was the case in <= v2.9
		# - As a single value, in which the case the minor version becomes 0
		# - As a float, which is seen as a major.minor version
		if u'API' not in d:
			d['API'] = StrictVersion(u'1.0')
		elif isinstance(d['API'], int):
			d['API'] = StrictVersion(u'%d.0' % d['API'])
		elif isinstance(d['API'], float):
			d['API'] = StrictVersion(str(d['API']))
		return d, s

	def generate_front_matter(self):

		"""
		desc:
			Generates YAML front matter. Any existing front matter is copied,
			except for the `OpenSesame`, `API`, and `Platform` fields, which are
			updated.

		returns:
			desc:	A front matter string.
			type:	str
		"""

		self.experiment.front_matter.update({
			'OpenSesame' : safe_str(metadata.__version__),
			'API' : metadata.api,
			'Platform' : os.name
			})
		front_matter = self.experiment.front_matter.copy()
		front_matter[u'API'] = metadata.api.version[0] \
			if metadata.api.version[1] == 0 \
			else float(str(metadata.api))
		return u'---\n%s---\n' % safe_decode(yaml.dump(
			front_matter, default_flow_style=False,
			allow_unicode=True))

	def split(self, s):

		"""
		desc:
			A unicode-safe bash-style split function. This is a wrapper around
			shlex.split() which is not unicode-safe on Python 2.

		arguments:
			s:
				desc:	The string to split.
				type:	[str, unicode]

		returns:
			desc:	The string split into a list.
			type:	list
		"""

		try:
			if py3:
				return shlex.split(s)
			# In Python 2, shlex is not unicode safe, so we need to do some manual
			# encoding and decoding
			return [safe_decode(_s) for _s in shlex.split(safe_encode(s))]
		except Exception as e:
			raise osexception(
				u'Failed to parse line "%s". Is there a closing quotation missing?' \
				% s, exception=e)

	def parse_cmd(self, cmd):

		"""
		desc:
			Parses OpenSesame command strings, which consist of a command,
			optionally followed by arguments, optionally followed by keywords.

			For example:

				widget 0 0 1 1 label text="Demo text"

			Here, `widget` is the command, `0` through `label` are arguments,
			and `text` is a keyword.

		arguments:
			cmd:
				desc:	The command string to parse.
				type:	[str, unicode]

		returns:
			desc:	A (command, arglist, kwdict) tuple.
			type:	tuple
		"""

		l = self.split(cmd)
		if len(l) == 0:
			return None, [], {}
		cmd = l[0]
		arglist = []
		kwdict = {}
		for s in l[1:]:
			m = self.re_cmd.match(s)
			if m is not None:
				arg = s[:m.end()-1]
				val = s[m.end():]
				kwdict[arg] = self.auto_type(val)
			else:
				arglist.append(self.auto_type(s))
		return cmd, arglist, kwdict

	def create_cmd(self, cmd, arglist=[], kwdict={}):

		"""
		desc:
			Generates a command string.

		arguments:
			cmd:
				desc:	The command.
				type:	str

		keywords:
			arglist:
				desc:	A list of compulsory order-specified arguments.
				type:	list
			kwdict:
				desc:	A dict of optional named keywords.
				type:	dict
		"""

		l = [cmd]
		for arg in arglist:
			l.append(self.safe_wrap(arg))
		for key, val in sorted(kwdict.items()):
			l.append(u'%s=%s' % (key, self.safe_wrap(val)))
		return u' '.join(l)

	def eval_text(self, txt, round_float=False, var=None):

		"""
		desc:
			Evaluates variables and inline Python in a text string.

			Examples:

				The resolution is [width] by [height] pixels
				This evaluates to 100: [=10x10]

		arguments:
			txt:	The string to evaluate. If the input is not a string, then
					the value will be returned unmodified.

		keywords:
			round_float:
				desc:	Indicates whether floating point values should be
						rounded or not.
				type:	bool

		returns:
			The evaluated string, or the input value for non-string input.
		"""

		def get_escape_sequence(m):
			return u'' if m.group(1) is None \
				else m.group(1)[:len(m.group(1))//2]

		if not isinstance(txt, basestring):
			return txt
		txt = safe_decode(txt)
		if var is None:
			var = self.experiment.var
		if round_float:
 			float_template = u'%%.%sf' % var.round_decimals
		while True:
			m = self.re_txt.search(txt)
			if m is None:
				break
			varname = m.group(2)[1:-1]
			val = var.get(varname)
			if round_float and isinstance(val, float):
				val = float_template % val
			else:
				val = safe_decode(val)
			txt = txt[:m.start(0)] + get_escape_sequence(m) + val \
				+ txt[m.end(0):]
		# Detect Python inlines [=10*10]
		while True:
			m = self.re_txt_py.search(txt)
			if m is None:
				break
			py = self.unescape(m.group(2)[2:-1])
			val = self.experiment.python_workspace._eval(py)
			txt = txt[:m.start(0)] + get_escape_sequence(m) + safe_decode(val) \
				+ txt[m.end(0):]
		return self.unescape(txt)

	def quotable_symbol(self, s):

		"""
		returns:
			desc:	True if s is a symbol that should be quoted in a conditional
					statement, and False otherwise.
			type:	bool
		"""

		# Don't quote operators
		if s in [u'not', u'or', 'and']:
			return False
		# Don't quote special keywords
		if s.lower() in [u'always', u'never']:
			return False
		# Don't quote numeric values
		try:
			int(s)
			return False
		except:
			return True

	def compile_cond(self, cnd, bytecode=True):

		"""
		desc:
			Compiles OpenSesame conditional statements.

			Examples:
				[width] > 100
				=var.width > 100

		arguments:
			cnd:
				desc:	The conditional statement to compile.
				type:	[str, unicode]

		keywords:
			bytecode:
				desc:	Indicates whether the conditional statement should be
						returned as bytecode (True) or a Python string (False).
				type:	bool

		returns:
			desc:	The conditional statement as a Python string or bytecode.
			type:	[str, bytecode]
		"""

		# Python conditions `=True` don't have to be evaluated
		if cnd.startswith(u'='):
			cnd = cnd[1:]
		else:
			# Quote all non-quoted symbols. This can be probably be done through
			# a clever regular expression, but for now we use a simple loop.
			while True:
				in_quote = None
				in_var = False
				symbol_start = None
				for i, ch in enumerate(cnd):
					# Don't scan within quoted strings
					if in_quote is None:
						if ch in u'"\'':
							in_quote = ch
							continue
					elif ch == in_quote:
							in_quote = None
							continue
					if in_quote:
						continue
					# Don't scan within variable definitions
					if ch == u'[':
						in_var = True
						continue
					elif ch == u']':
						in_var = False
						continue
					if in_var:
						continue
					# Detect symbols starts, i.e. the first alphanumeric 
					# character or underscore
					if ch.isalnum() or ch == u'_':
						if symbol_start is None:
							symbol_start = i
					# Detect symbol ends, i.e. the first non-alphanumeric 
					# character after an alphanumeric character or an 
					# underscore.
					elif symbol_start is not None:
						symbol = cnd[symbol_start:i]
						symbol_range = symbol_start, i
						symbol_start = None
						if not self.quotable_symbol(symbol):
							continue
						cnd = cnd[:symbol_range[0]] + u'"' + symbol + u'"' \
							+ cnd[symbol_range[1]:]
						break
				else:
					# The for-else clause happens when the for loop was not
					# broken. If no break occurred, then nothing was quoted,
					# and we don't need to restart the for loop, i.e. we can
					# break the infinite while. We only need to check that there
					# was no symbol still being processed, in which case it
					# needs to be quoted.
					if symbol_start is not None:
						symbol = cnd[symbol_start:]
						if self.quotable_symbol(symbol):
							cnd = cnd[:symbol_start] + u'"' + symbol + u'"'
					break
			# Replace [variables] by var.variables
			while True:
				m = self.re_txt.search(cnd)
				if m is None:
					break
				cnd = cnd[:m.start()] + u'var.%s' % m.group()[1:-1] \
					+ cnd[m.end():]
			# Replace single equals signs (=) by doubles (==)
			cnd = self.re_single_eq.sub(u'==', cnd)
			# Replace always and never words by True or False
			cnd = self.re_never.sub(u'False', cnd)
			cnd = self.re_always.sub(u'True', cnd)
		if bytecode:
			try:
				return compile(cnd, u"<conditional statement>", u"eval")
			except:
				raise osexception(
					u"'%s' is not a valid conditional statement" % cnd)
		return self.unescape(cnd)

	def unescape(self, s):

		"""
		desc:
			Unescapes square brackets in text.

		arguments:
			s:
				desc:	The text to unescape.
				type:	str

		returns:
			desc:	The unescaped text.
			type:	str
		"""

		s = safe_decode(s)
		return s.replace(u'\[', u'[').replace(u'\]', u']')

	def safe_wrap(self, s):

		"""
		desc:
			Wraps and escapes a text so that it can safely be embedded in a
			command string. For example:

			He said: "hi"

			would become:

			"He said: \"hi\""

		arguments:
			s:
				desc:	The text to wrap.
				type:	str

		returns:
			desc:	The wrapped text.
			type:	str
		"""

		s = safe_decode(s).replace(u'\\', u'\\\\')
		try:
			float(s)
			assert(s == s.strip())  # No flanking whitespace
		except:
			if not s.replace(u'_', u'').isalnum():
				return u'"%s"' % s.replace(u'"', u'\\"')
		return s


	def sanitize(self, s, strict=False, allow_vars=True):

		"""
		desc:
			Removes invalid characters (notably quotes) from the string.

		arguments:
			s:
				desc:	The string to be sanitized. This can be any type, but
						if it is not unicode, it will be coerced to unicode.

		keywords:
			strict:
				desc:	If True, all except underscores and alphanumeric
						characters are stripped.
				type:	bool
			allow_vars:
				desc:	If True, square brackets are not sanitized, so you can
						use variables.
				type:	bool

		returns:
			desc:	A sanitized string.
			type:	unicode

		example: |
			# Prints 'Universit Aix-Marseille'
			print(self.sanitize('\"Université Aix-Marseille\"'))
			# Prints 'UniversitAixMarseille'
			print(self.sanitize('\"Université Aix-Marseille\""', strict=True))
		"""

		s = safe_decode(s)
		if strict:
			if allow_vars:
				return self.re_sanitize_strict_vars.sub(u'', s)
			return self.re_sanitize_strict_novars.sub(u'', s)
		return self.re_sanitize_loose.sub(u'', s)

	def to_ascii(self, s, strict=False):

		"""
		desc:
			Converts all non-ASCII characters to U+XXXX notation, so that the
			resulting string can be treated as plain ASCII text.

		arguments:
			s:
				desc:	A unicode string to be santized
				type:	unicode

		keywords:
			strict:
				desc:	If True, special characters are ignored rather than
						recoded.
				type:	bool

		returns:
			desc:	A regular Python string with all special characters replaced
					by U+XXXX notation or ignored (if strict).
			type:	str
		"""

		if strict:
			_s = safe_encode(s, enc=u'ascii', errors=u'ignore')
		else:
			_s = codecs.encode(s, u'ascii', u'osreplace')
		_s = safe_decode(_s)
		return _s.replace(os.linesep, u'\n')

	def from_ascii(self, s):

		"""
		desc:
			Converts an ascii str with U+XXXX notation to actual Unicode.

		arguments:
			s:
			 	desc:	A plain-ascii string.
				type:	str

		returns:
			desc:	A string with all U+XXXX characters converted to the corresponding
					unicode characters.
			type:	unicode
		"""

		if not isinstance(s, basestring):
			raise osexception(
				u'from_ascii() expects first argument to be unicode or str, not "%s"' \
				% type(s))
		s = safe_decode(s)
		while True:
			m = self.re_from_ascii.search(s)
			if m is None:
				break
			if py3:
				_unichr = chr
			else:
				_unichr = unichr
			s = s.replace(m.group(0), _unichr(int(m.group(1), 16)), 1)
		return s

	def valid_var_name(self, s):

		"""
		desc:
			Checks whether a string is a valid variable name.

		arguments:
			s:
				desc:	The string to check.
				type:	str

		returns:
			type:	True if s is a variable valid name, False if not.
		"""

		return re.match(self.re_valid_var_name, s)

def osreplace(exc):

	"""
	desc:
		A replacement function to allow opensame-style replacement of unicode
		characters.

	arguments:
		exc:
			type:	UnicodeEncodeError

	returns:
		desc:	A (replacement, end) tuple.
		type:	tuple
	"""

	_s = u''
	for ch in exc.object[exc.start:exc.end]:
		_s += u'U+%.4X' % ord(ch)
	return _s, exc.end

codecs.register_error(u'osreplace', osreplace)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *

from libopensesame.exceptions import osexception

def float_list(l, desc, min_len=None, max_len=None):
	
	"""
	Converts a variable to a list of floats if possible.
	
	Arguments:
	a		--	The variable to convert.
	desc	--	A description to clarify the osexception.
	
	Keyword arguments:
	min_len	--	The minimum length of the list. (default=None)
	max_len	--	The maximum length of the list. (default=None)
	
	Raises:
	A osexception if the variable could not be converted.
	
	Returns:
	A list of floats.
	"""
		
	try:
		l = list(l)
	except:
		raise osexception( \
			u'Expecting a list or compatible type not "%s" for "%s"' % (l, \
				desc))
	if min_len is not None and len(l) < min_len:
		raise osexception( \
			u'Expecting a list of at least %d items for "%s"' % (min_len, desc))
	if max_len is not None and len(l) > max_len:
		raise osexception( \
			u'Expecting a list of at most %d items for "%s"' % (max_len, desc))
	return l


#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import platform
import pygame
from pygame.locals import *
from string import whitespace
from libopensesame.exceptions import osexception
from openexp._keyboard.keyboard import Keyboard
from openexp.backend import configurable

# Whitespace, backspace, and empty strings are not acceptable names for keys.
# These should be converted to descriptions, e.g. '\t' to 'tab'
invalid_unicode = [u'', u'\x08', u'\x7f'] + list(whitespace)
# On mac arrow keys are not accepted as valid input either. Add them to this list
if platform.system() == "Darwin":
	invalid_unicode += [
		u'\uf702', # left
		u'\uf703', # right
		u'\uf700', # up
		u'\uf701', # down
		u'\uf729', # home
		u'\uf72b', # end
		u'\uf72c', # page up
		u'\uf72d', # page down
		u'\uf728', # delete
		u'\uf739', # numlock
	]


class Legacy(Keyboard):

	"""
	desc:
		This is a keyboard backend built on top of PyGame.
		For function specifications and docstrings, see
		`openexp._keyboard.keyboard`.
	"""

	def __init__(self, experiment, **resp_args):

		pygame.init()
		self.key_code_to_name = {}
		self.key_name_to_code = {}
		for i in dir(pygame):
			if i[:2] == u"K_":
				code = getattr(pygame, i)
				name1 = self.key_name(code).lower()
				name2 = name1.upper()
				name3 = i[2:].lower()
				name4 = name3.upper()
				self.key_code_to_name[code] = [name1, name2, name3, name4]
				try:
					i = int(name5)
					self.key_code_to_name[code].append(name5)
				except:
					pass
				self.key_name_to_code[name1] = code
				self.key_name_to_code[name2] = code
				self.key_name_to_code[name3] = code
				self.key_name_to_code[name4] = code
		self.persistent_virtual_keyboard = False
		Keyboard.__init__(self, experiment, **resp_args)

	@configurable
	def get_key(self):

		return self._get_key_event(pygame.KEYDOWN)

	@configurable
	def get_key_release(self):

		return self._get_key_event(pygame.KEYUP)

	def _get_key_event(self, event_type):

		start_time = pygame.time.get_ticks()
		time = start_time
		keylist = self.keylist
		timeout = self.timeout
		while True:
			time = pygame.time.get_ticks()
			# Some input methods send multiple key events at the same time, 
			# for example when composing a multicharacter Chinese or Japanese 
			# string. That's why we process up all events, rather than 
			# assuming that there's only a single relevant event in the queue.
			key = u''
			for event in pygame.event.get(event_type):
				if event.key == pygame.K_ESCAPE:
					self.experiment.pause()
				# KEYUP events don't have a unicode property, so in that case
				# we fall back to converting the key code straight to an ASCII
				# value. This is not great, because it assumes a QWERTY
				# keyboard layout.
				if hasattr(event, u'unicode'):
					ucode = event.unicode
				elif event.key < 128:
					ucode = chr(event.key)
				else:
					ucode = u''
				if ucode in invalid_unicode:
					key += self.key_name(event.key)
				else:
					key += ucode
			if key and (keylist is None or key in keylist):
				return key, time
			if timeout is not None and time - start_time >= timeout:
				break
		return None, time

	def get_mods(self):

		l = []
		mods = pygame.key.get_mods()
		if mods & KMOD_LSHIFT or mods & KMOD_RSHIFT or mods & KMOD_SHIFT:
			l.append(u"shift")
		if mods & KMOD_LCTRL or mods & KMOD_RCTRL or mods & KMOD_CTRL:
			l.append(u"ctrl")
		if mods & KMOD_LALT or mods & KMOD_RALT or mods & KMOD_ALT:
			l.append(u"alt")
		if mods & KMOD_LMETA or mods & KMOD_RMETA or mods & KMOD_META:
			l.append(u"meta")
		return l

	def valid_keys(self):

		return sorted(self.key_name_to_code.keys())

	def synonyms(self, key):

		# If the key is not familiar, simply return it plus its string
		# representation.
		if key not in self.key_name_to_code:
			return [key, safe_decode(key)]
		return self.key_code_to_name[self.key_name_to_code[key]]

	def flush(self):

		keypressed = False
		for event in pygame.event.get():
			if event.type == KEYDOWN:
				keypressed = True
				if event.key == pygame.K_ESCAPE:
					self.experiment.pause()
		pygame.event.pump()
		return keypressed

	def key_name(self, key):

		return str(pygame.key.name(key)).replace(u'[', u'').replace(u']',
			u'')

	def _keycode_to_str(self, keycode):

		return safedecode(pygame.key.name(keycode)) \
			.replace(u'[', u'').replace(u']', u'').lower()


# Non PEP-8 alias for backwards compatibility
legacy = Legacy



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import sys
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from pyqode.core import panels, modes
from pyqode.core.backend import server
from pyqode.core.api import (
	CodeEdit,
	Panel,
	CharBasedFoldDetector,
	IndentFoldDetector,
	ColorScheme
)
from pyqode_extras.modes import AutodetectIndentationMode

# For markdown it's annoying if quotes are autocompleted
DISABLE_AUTO_COMPLETE_QUOTES = ['markdown']


class FallbackCodeEdit(CodeEdit):

	# generic
	mimetypes = []

	#: the list of mimetypes that use char based fold detector
	_char_based_mimetypes = [
		'text/x-php',
		'text/x-c++hdr',
		'text/x-c++src',
		'text/x-chdr',
		'text/x-csrc',
		'text/x-csharp',
		'text/x-plain',
		'application/javascript'
	]

	def __init__(self, parent=None):

		super(FallbackCodeEdit, self).__init__(parent)
		self._start_backend()
		self.show_whitespaces = cfg.pyqode_show_whitespaces
		if cfg.pyqode_image_annotations:
			self.panels.append(
				panels.ImageAnnotationsPanel(),
				panels.ImageAnnotationsPanel.Position.RIGHT
			)
		if cfg.pyqode_show_line_numbers:
			self.panels.append(panels.LineNumberPanel())
		self.panels.append(
			panels.SearchAndReplacePanel(),
			Panel.Position.BOTTOM
		)
		if cfg.pyqode_code_folding:
			self.panels.append(panels.FoldingPanel())
		self.modes.append(modes.CursorHistoryMode())
		self.modes.append(modes.ExtendedSelectionMode())
		self.modes.append(modes.CaseConverterMode())
		self.modes.append(modes.LineSorterMode())
		self.modes.append(modes.FileWatcherMode())
		if cfg.pyqode_highlight_caret_line:
			self.modes.append(modes.CaretLineHighlighterMode())
		self.modes.append(modes.PygmentsSyntaxHighlighter(
			self.document(),
			color_scheme=ColorScheme(cfg.pyqode_color_scheme))
		)
		self.modes.append(modes.ZoomMode())
		self.modes.append(modes.AutoIndentMode())
		self.modes.append(modes.IndenterMode())
		self.modes.append(modes.SymbolMatcherMode())
		self.modes.append(modes.SmartBackSpaceMode())
		self.modes.append(AutodetectIndentationMode())
		self.panels.append(panels.EncodingPanel(), Panel.Position.TOP)
		self.panels.append(panels.ReadOnlyPanel(), Panel.Position.TOP)
		if cfg.pyqode_fixed_width:
			self.panels.append(
				panels.MarginPanel(nchar=cfg.pyqode_fixed_width_nchar),
				Panel.Position.LEFT
			)
		if cfg.pyqode_right_margin:
			rmm = modes.RightMarginMode()
			self.modes.append(rmm)
			try:
				rmm.color = self.syntax_highlighter.formats['comment'] \
					.foreground().color()
			except Exception as e:  # Don't know how safe this is
				pass
		self._set_comment_mode()
		
	def _start_backend(self):
		
		self._backend.start(
			server.__file__,
			sys.executable,
			reuse=True
		)
		
	def _set_comment_mode(self):
		
		comment_prefix = self.comment_prefix
		if comment_prefix:
			if 'CommentsMode' in self.modes.keys():
				self.modes.get('CommentsMode').prefix = comment_prefix
			else:
				self.modes.append(modes.CommentsMode(prefix=comment_prefix))
		elif 'CommentsMode' in self.modes.keys():
			self.modes.remove('CommentsMode')

	def setPlainText(self, txt, mime_type='', encoding=''):

		if not mime_type:
			mime_type = self.file.mimetype
		if not encoding:
			encoding = self.file.encoding
		if mime_type != u'text/x-plain':
			try:
				self.syntax_highlighter.set_lexer_from_mime_type(mime_type)
			except Exception:
				# The syntax_highlighter may not exist, or there may not be a lexer
				# for the mime type
				pass
			else:
				if mime_type in self._char_based_mimetypes:
					self.syntax_highlighter.fold_detector = CharBasedFoldDetector()
				else:
					self.syntax_highlighter.fold_detector = IndentFoldDetector()
		if self.language in DISABLE_AUTO_COMPLETE_QUOTES:
			self._disable_auto_complete_for_quotes()
		else:
			self._enable_auto_complete_for_quotes()
		super(FallbackCodeEdit, self).setPlainText(txt, mime_type, encoding)
		if (
			cfg.pyqode_code_completion and
			self.language not in cfg.pyqode_code_completion_excluded_mimetypes
		):
			self.modes.append(modes.AutoCompleteMode())
			self.modes.append(modes.CodeCompletionMode())
			# The AutoIndentMode should catch key presses after the
			# CodeCompletionMode, otherwise code completions will have an ugly
			# newline in it. Toggling the state accomplishes this, although
			# it's an ugly hack.
			if 'AutoIndentMode' in self.modes.keys():
				auto_indent_mode = self.modes.get('AutoIndentMode')
				if auto_indent_mode.enabled:
					auto_indent_mode.enabled = False
					auto_indent_mode.enabled = True
		self._set_comment_mode()

	def _enable_auto_complete_for_quotes(self):

		try:
			auto_complete_mode = self.modes.get('AutoCompleteMode')
		except KeyError:
			oslogger.debug('AutoCompleteMode not enabled')
			return
		auto_complete_mode.MAPPING['\''] = '\''
		auto_complete_mode.MAPPING['"'] = '"'

	def _disable_auto_complete_for_quotes(self):

		try:
			auto_complete_mode = self.modes.get('AutoCompleteMode')
		except KeyError:
			oslogger.debug('AutoCompleteMode not enabled')
			return
		if '\'' not in auto_complete_mode.MAPPING:
			return
		del auto_complete_mode.MAPPING['\'']
		del auto_complete_mode.MAPPING['"']

	def clone(self):

		clone = self.__class__(parent=self.parent())
		return clone

	@property
	def language(self):

		filenames = self.syntax_highlighter._lexer.filenames
		if u'*.R' in filenames:
			return u'R'
		if u'*.md' in filenames:
			return u'markdown'
		if u'*.js' in filenames:
			return u'javascript'
		return u'unknown'
		
	@property
	def comment_prefix(self):
		
		lang = self.language
		if lang == 'R':
			return '# '
		if lang == 'javascript':
			return '// '

	def __repr__(self):

		return 'FallbackCodeEdit(path={}, language={})'.format(
			self.file.path,
			self.language
		)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import platform
from openexp._canvas._richtext.richtext import RichText
from openexp._canvas._element.psycho import PsychoElement
from psychopy import visual


class Psycho(PsychoElement, RichText):

	def prepare(self):

		im = self._to_pil()
		if platform.system() == 'Darwin':
			# When displaying on Mac retina screens, the resolutions reported
			# by psychopy's window object and OpenSesame's experiment object
			# may diverge. This results in incorrectly rendered text with
			# respect to scale and positioning. We correct for the discrepancy
			# between reported sizes by the ratios calculated below. If the
			# sizes reported by win and experiment are equal, the ratio values
			# should end up as 1, and nothing changes.
			x_ratio = int(self.win.size[0] / self.experiment.width)
			y_ratio = int(self.win.size[1] / self.experiment.height)
			# Only resize if necessary to prevent unnecessary operations
			if x_ratio != 1 or y_ratio != 1:
				im = im.resize((im.width * x_ratio, im.height * y_ratio))
			x, y = self.to_xy(self.x * x_ratio, self.y * y_ratio)
		else:
			x, y = self.to_xy(self.x, self.y)
		if not self.center:
			x += im.width // 2
			y -= im.height // 2
		self._stim = visual.SimpleImageStim(self.win, im, pos=(x, y))



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.sketchpad import sketchpad as sketchpad_runtime
from libqtopensesame.items.qtplugin import qtplugin
from libqtopensesame.items.feedpad import feedpad
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'sketchpad', category=u'item')

class sketchpad(feedpad, qtplugin, sketchpad_runtime):

	"""
	desc:
		The sketchpad controls are implemented in feedpad.
	"""

	description = _(u'Displays stimuli')

	def __init__(self, name, experiment, string=None):

		sketchpad_runtime.__init__(self, name, experiment, string)
		qtplugin.__init__(self)

	def init_edit_widget(self):

		"""
		desc:
			Initializes the widget.
		"""

		feedpad.init_edit_widget(self)
		self.sketchpad_widget.ui.widget_settings_reset_variables.hide()



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._canvas._richtext.richtext import RichText
from openexp._canvas._element.legacy import LegacyElement
import pygame


class Legacy(LegacyElement, RichText):

	def prepare(self):

		if not hasattr(self, '_text_surface') or self._dirty:
			im = self._to_pil()
			self._text_surface = pygame.image.fromstring(
				im.tobytes(), im.size, im.mode)
			self._dirty = False
		x, y = self.to_xy(self.x, self.y)
		if self.center:
			x -= self._text_surface.get_width()//2
			y -= self._text_surface.get_height()//2
		self.surface.blit(self._text_surface, (x, y))

	@staticmethod
	def _setter(key, self, val):

		self._dirty = True
		RichText._setter(key, self, val)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._log.log import Log
import os


class Csv(Log):

	"""
	desc:
		For docstrings, see openexp._log.log.
	"""

	def __init__(self, experiment, path):

		self._log = None
		Log.__init__(self, experiment, path)

	def close(self):

		if self._log is not None:
			self._log.close()

	def open(self, path):

		if self._log is not None:
			self.close()
		# If only a filename is present, we interpret this filename as relative
		# to the experiment folder, instead of relative to the current working
		# directory.
		if (
			os.path.basename(path) == path and
			self.experiment.experiment_path is not None
		):
			self._path = os.path.join(self.experiment.experiment_path, path)
		else:
			self._path = path
		# Open the logfile
		self.experiment.var.logfile = self._path
		if self._path not in self.experiment.data_files:
			self.experiment.data_files.append(self._path)
		self._log = safe_open(self._path, u'w')
		self._header_written = False

	def write(self, msg, newline=True):

		self._log.write(safe_decode(msg))
		if newline:
			self._log.write(u'\n')
		# Flush to avoid pending write operations
		self._log.flush()
		os.fsync(self._log)

	def write_vars(self, var_list=None):

		if var_list is None:
			var_list = self.all_vars()
		if not self._header_written:
			l = [u'"%s"' % var.replace(u'"', u'\\"') for var in var_list]
			self.write(u','.join(l))
			self._header_written = True
		l = []
		for var in var_list:
			val = self.experiment.var.get(var, _eval=False, default=u'NA')
			val = safe_decode(val)
			l.append(u'"%s"' % val.replace(u'"', u'\\"'))
		self.write(u','.join(l))


# Non PEP-8 alias for backwards compatibility
csv = Csv



#!/usr/bin/env python
#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import sys
import shutil
from setuptools import setup
from libopensesame import metadata
from setup_shared import included_plugins, included_extensions
import fnmatch

# Fix a bug in the packaging of stdeb
try:
	from stdeb3 import util
except ImportError:
	pass
else:
	if not hasattr(util, 'RULES_BINARY_ALL_TARGET'):
		util.RULES_BINARY_ALL_TARGET = '\nbinary: binary-indep\n'

# For debian packaging, we use a slightly different version-number scheme, which
# is triggered by setting the USEDEBVERSION environment variable.
if 'USEDEBVERSION' in os.environ and os.environ['USEDEBVERSION'] == '1':
	version = metadata.deb_version
else:
	version = metadata.__version__

EXCLUDE = [
	u'[\\/].*',
	u'*eco_alt_template.osexp',
	u'*~',
	u'*.pyc',
	u'*.pyo',
	u'*__pycache__*'
	]


def is_excluded(path):

	for m in EXCLUDE:
		if fnmatch.fnmatch(path, m):
			return True
	return False


def resources():

	"""
	desc:
		Create a list of all resource files that need to be included

	returns:
		A list of (target folder, filenames) tuples.
	"""

	l = []
	for root, dirnames, filenames in os.walk(u'opensesame_resources'):
		for f in filenames:
			path = os.path.join(root, f)
			if not is_excluded(path):
				l.append((os.path.join(u'share', root), [path]))
	return l


def recursive_glob(src_folder, target_folder):

	"""
	desc:
		Recursively gets all files that are in src folder.

	arguments:
		src_folder:		The source folder.
		target_folder:	The target folder.

	returns:
		A list of (target folder, filenames) tuples.
	"""

	l = []
	print(src_folder)
	path_list = []
	for path in os.listdir(src_folder):
		full_path = os.path.join(src_folder, path)
		if os.path.isdir(full_path):
			l += recursive_glob(full_path, os.path.join(target_folder, path))
			continue
		if is_excluded(full_path):
			continue
		print(u'\t%s' % full_path)
		path_list.append(full_path)
	l.append( (os.path.join(u'share', target_folder), path_list) )
	return l


def plugins(included, _type=u'plugins'):

	"""
	desc:
		Create a list of all plugin files that need to be included.

	arguments:
		included:	A list of included plugin names.
		_type:		The plugin type, i.e. 'plugins' or 'extensions'.

	returns:
		A list of (target folder, filenames) tuples.
	"""

	l = []
	for plugin in os.listdir(u'opensesame_%s' % _type):
		if plugin in included:
			folder = os.path.join(u'opensesame_%s' % _type, plugin)
			l += recursive_glob(folder, folder)
	return l


def data_files():

	return (
		[
			(u"share/icons/hicolor/scalable/apps", [u"mime/opensesame.svg"]),
			(u"share/applications", [u"mime/opensesame.desktop"]),
		] +
		plugins(included=included_plugins, _type=u'plugins') +
		plugins(included=included_extensions, _type=u'extensions') +
		resources()
	)


# Temporarily create README.txt
shutil.copy(u'readme.md', u'README.txt')
with open('readme.md') as fd:
	readme = fd.read()
setup(
	# The PyPi name is python-opensesame
	name=u'opensesame' if u'bdist_deb' in sys.argv else u'python-opensesame',
	version=version,
	description=u"A graphical experiment builder for the social sciences",
	long_description=readme,
	long_description_content_type='text/markdown',
	author=u"Sebastiaan Mathot",
	author_email=u"s.mathot@cogsci.nl",
	url=u"http://osdoc.cogsci.nl/",
	entry_points={
		'gui_scripts': [
			'opensesame = libqtopensesame.__main__:opensesame',
			'opensesamerun = libqtopensesame.__main__:opensesamerun',
		]
	},
	# scripts=['opensesame', 'opensesamerun'],
	include_package_data=False,
	packages=[
		"openexp",
		"openexp._canvas",
		"openexp._canvas._arrow",
		"openexp._canvas._circle",
		"openexp._canvas._element",
		"openexp._canvas._ellipse",
		"openexp._canvas._fixdot",
		"openexp._canvas._gabor",
		"openexp._canvas._image",
		"openexp._canvas._line",
		"openexp._canvas._noise_patch",
		"openexp._canvas._polygon",
		"openexp._canvas._rect",
		"openexp._canvas._richtext",
		"openexp._keyboard",
		"openexp._mouse",
		"openexp._sampler",
		"openexp._coordinates",
		"openexp._clock",
		"openexp._log",
		"openexp._color",
		"libopensesame",
		"libopensesame.osexpfile",
		"libopensesame.widgets",
		"libopensesame.widgets.themes",
		"libopensesame.sketchpad_elements",
		"libqtopensesame",
		"libqtopensesame.extensions",
		"libqtopensesame.actions",
		"libqtopensesame.dialogs",
		"libqtopensesame.items",
		"libqtopensesame.misc",
		"libqtopensesame.runners",
		"libqtopensesame.widgets",
		"libqtopensesame.validators",
		"libqtopensesame.sketchpad_elements",
		"libqtopensesame._input",
		],
	data_files=data_files(),
	install_requires=[
		'pyqode3.python',
		'python-qdatamatrix',
		'python-pseudorandom',
		'python-qnotifications',
		'python-fileinspector',
		'PyYAML',
		'webcolors',
		],
	classifiers=[
		'Development Status :: 4 - Beta',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering',
		'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 3',
		],
	)

# Clean up temporary readme
os.remove(u'README.txt')



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_subcomponent import base_subcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tree_base_item', category=u'core')

class tree_base_item(base_subcomponent, QtWidgets.QTreeWidgetItem):

	"""
	desc:
		A base class that corresponds to any widget in the overview area.
	"""

	def start_edit(self, col):

		self.treeWidget().editItem(self, col)

	def ancestry(self):

		return None

	def path(self):

		if self.parent() is None:
			return self.text(0)
		return self.parent().path() + u'.' + self.text(0)

	def children(self):

		l = []
		for i in range(self.childCount()):
			child = self.child(i)
			l += [child] + child.children()
		return l

	def show_context_menu(self, pos):

		pass

	def paste(self):

		pass

	def copy(self):

		pass

	def create_linked_copy(self):

		pass

	def create_unlinked_copy(self):

		pass

	def copy_linked(self):

		pass

	def copy_unlinked(self):

		pass

	def delete(self):

		pass

	def permanently_delete(self):

		pass

	def open_tab(self):

		pass

	def drop_hint(self):

		return None

	def droppable(self, data):

		return self._droppable

	def set_droppable(self, droppable):

		self._droppabe = droppable

	def draggable(self):

		return self._draggable

	def set_draggable(self, draggable):

		self._draggable = draggable

	def has_append_menu(self):

		return False

	def expand(self):

		"""
		desc:
			Expands this item and all items under it.
		"""

		self.setExpanded(True)
		for i in range(self.childCount()):
			self.child(i).expand()

	def collapse(self):

		"""
		desc:
			Collapses this item and all items under it.
		"""

		self.setExpanded(False)
		for i in range(self.childCount()):
			self.child(i).collapse()

	def rename(self, from_name, to_name):

		"""
		desc:
			Renames an item.

		arguments:
			from_name:
				desc:	The old name.
				type:	unicode
			to_name:
				desc:	The new name.
				type:	unicode
		"""

		for i in range(self.childCount()):
			self.child(i).rename(from_name, to_name)

	def set_icon(self, name, icon):

		"""
		desc:
			Changes an item's icon.

		arguments:
			name:
				desc:	The item name.
				type:	unicode
			icon:
				desc:	The icon name.
				type:	unicode
		"""

		for i in range(self.childCount()):
			self.child(i).set_icon(name, icon)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame.base_response_item import base_response_item
from openexp.keyboard import keyboard


class keyboard_response_mixin(object):

	"""
	desc:
		A mixin class that should be inherited along with base_response_item
		by all classes that want to collect keyboard responses.
	"""

	def prepare_response_func(self):

		"""See base_response_item."""

		self._keyboard = keyboard(self.experiment, timeout=self._timeout,
			keylist=self._allowed_responses)
		if self.var.get(u'event_type', default=u'keypress') == u'keypress':
			return self._keyboard.get_key
		if self.var.event_type == u'keyrelease':
			return self._keyboard.get_key_release
		raise osexception(u'event_type should be "keypress" or "keyrelease"')


class keyboard_response(keyboard_response_mixin, base_response_item):

	"""
	desc:
		An item for collecting keyboard responses.
	"""

	description = u'Collects keyboard responses'
	process_feedback = True

	def reset(self):

		"""See item."""

		self.var.flush = u'yes'
		self.var.timeout = u'infinite'
		self.var.duration = u'keypress'
		self.var.event_type = u'keypress'
		self.var.unset(u'allowed_responses')
		self.var.unset(u'correct_response')

	def prepare(self):

		"""See item."""

		base_response_item.prepare(self)
		self._flush = self.var.flush == u'yes'

	def response_matches(self, test, ref):

		"""See base_response_item."""

		return any(safe_decode(r) in self._keyboard.synonyms(test) for r in ref)

	def run(self):

		"""See item."""

		if self._flush:
			self._keyboard.flush()
		base_response_item.run(self)

	def coroutine(self):

		"""See coroutines plug-in."""

		self._keyboard.timeout = 0
		alive = True
		yield
		self._t0 = self.set_item_onset()
		if self._flush:
			self._keyboard.flush()
		while alive:
			key, time = self._keyboard.get_key()
			if key is not None:
				break
			alive = yield
		self.process_response((key, time))



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from openexp.sampler import Sampler
try:
	import numpy as np
	from scipy import signal
except:
	np = None
	signal = None


def Synth(experiment, osc="sine", freq=440, length=100, attack=0, decay=5):

	"""
	desc: |
		A factory that synthesizes a sound and returns it as a `sampler object`.
		For a full description of keywords, see `python_workspace_api.synth`.

		For backwards compatibility, this function behaves as though it is a
		back-end.

	arguments:
		experiment:
			desc:		The experiment object.
			type:		experiment

	returns:
		desc:	A SAMPLER object.
		type:	sampler
	"""

	if np is None:
		raise osexception(
			u'The synth is not available, because numpy is missing.')
	if attack < 0 or attack > length:
		raise osexception(
			u'Attack must be a numeric value between 0 and the sound length')
	if decay < 0 or decay > length:
		raise osexception(
			u'Decay must be a numeric value between 0 and the sound length')
	# We need to multiply the rate by two to get a stereo signal
	rate = 2*experiment.var.get(u'sound_freq', 48000)
	signal = osc_gen(osc, key_to_freq(freq), length, rate)
	_envelope = envelope(length, attack, decay, rate)
	sound = to_int_16(signal * _envelope)
	return Sampler(experiment, sound)


def key_to_freq(key):

	"""
	desc:
		Converts a key (e.g., A1) to a frequency.

	arguments:
		key:
			desc:	A string like "A1", "eb2", etc, or a numeric frequency
					(in which case the frequency is simply returned as a
					float).
			type:	[str, unicode, int, float]

	returns:
		desc:		A frequency in hertz.
		type:		float

	example: |
		from openexp.synth import synth
		my_synth = synth(exp)
		print('An a2 is %d Hz' % my_synth.key_to_freq('a2'))
	"""

	if type(key) in [int, float]:
		return key
	if not isinstance(key, basestring) or len(key) < 2:
		raise osexception(
			"synth.key_to_freq(): '%s' is not a valid note, expecting something like 'A1'")
	n = key[:-1].lower()
	try:
		o = int(key[-1])
	except:
		raise osexception(
			"synth.key_to_freq(): '%s' is not a valid note, expecting something like 'A1'")
	if n == "a":
		f = 440.0
	elif n == "a#" or n == "bb":
		f = 466.16
	elif n == "b":
		f = 493.92
	elif n == "c":
		f = 523.28
	elif n == "c#" or n == "db":
		f = 554.40
	elif n == "d":
		f = 587.36
	elif n == "d#" or n == "eb":
		f = 698.47
	elif n == "e":
		f = 659.48
	elif n == "f":
		f = 698.48
	elif n == "f#" or n == "gb":
		f = 740.00
	elif n == "g":
		f = 784.00
	elif n == "ab" or n == "g#":
		f == 830.64
	if o < 1:
		o = 0.5 ** (abs(o) + 1)
		freq = f * o
	else:
		freq = f * o
	return freq


def osc_gen(_type, freq, length, rate):

	length *= .001
	t = np.linspace(0, length, int(length*rate))
	if _type == u'square':
		return signal.square(2 * np.pi * freq * t)
	if _type == u'saw':
		return signal.sawtooth(2 * np.pi * freq * t)
	if _type == u'sine':
		return np.sin(2 * np.pi * freq * t)
	if _type == u'white_noise':
		return np.random.random(int(length * rate)) * 2 - 1
	raise osexception(u'Invalid oscillator: %s' % _type)


def envelope(length, attack, decay, rate):

	"""
	desc:
		An envelope generator that determines the volume profile of the
		sound.

	visible:
		False
	"""

	length *= .001
	attack *= .001
	decay *= .001
	e = np.ones(int(length*rate))
	if attack > 0:
		attack = int(attack*rate)
		e[:attack] *= np.linspace(0, 1, attack)
	if decay > 0:
		decay = int(decay*rate)
		e[-decay:] *= np.linspace(1, 0, decay)
	return e


def to_int_16(a):

	"""
	desc:
		Converts the float array to an 16 bit int array, which is a more
		typical sound format.

	visible:
		False
	"""

	a *= 32767
	return a.astype(np.int16)


# Non PEP-8 alias for backwards compatibility
synth = Synth



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import metadata
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_component import base_component
from libqtopensesame.misc import theme

class qtopensesamerun(QtWidgets.QMainWindow, base_component):

	"""Implements the GUI for opensesamerun."""

	def __init__(self, options, parent=None):

		"""
		Constructor.

		Arguments:
		options		--	Command-line arguments passed to opensesamerun, as
						parsed by `libopensesame.misc.opensesamerun_options()`.

		Keyword arguments:
		parent		--	A parent QWidget. (default=None)
		"""

		# Construct the parent
		QtWidgets.QMainWindow.__init__(self, parent)
		# Setup the UI
		self.load_ui(u'misc.opensesamerun')
		self.ui.button_run.clicked.connect(self.run)
		self.ui.label_header.setText(
			self.ui.label_header.text() % metadata.__version__)
		self.theme = theme.theme(self)
		self.ui.button_browse_experiment.clicked.connect(self.browse_experiment)
		self.ui.button_browse_logfile.clicked.connect(self.browse_logfile)
		self.options = options
		# Fill the GUI controls based on the options
		self.ui.edit_experiment.setText(self.options.experiment)
		self.ui.checkbox_fullscreen.setChecked(self.options.fullscreen)
		self.ui.checkbox_pylink.setChecked(self.options.pylink)
		self.ui.spinbox_subject_nr.setValue(int(self.options.subject))
		self.ui.edit_logfile.setText(self.options.logfile)

	def browse_experiment(self):

		"""Locates the experiment file."""

		file_type_filter = \
			u"OpenSesame files (*.osexp *.opensesame.tar.gz *.opensesame);;OpenSesame script and file pool (*.opensesame.tar.gz);;OpenSesame script (*.opensesame)"
		path = QtWidgets.QFileDialog.getOpenFileName(self, \
			u"Open experiment file", filter = file_type_filter)
		# In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
		# a string, of which the first position contains the path.
		if isinstance(path,tuple):
			path = path[0]
		if path == u"":
			return
		self.ui.edit_experiment.setText(path)

	def browse_logfile(self):

		"""Locates the logfile.	"""

		path = QtWidgets.QFileDialog.getSaveFileName(self, \
			u"Choose a location for the logfile")
		# In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
		# a string, of which the first position contains the path.
		if isinstance(path,tuple):
			path = path[0]

		if path == u"":
			return
		self.ui.edit_logfile.setText(path)

	def show(self):

		"""Sets the run flag to false."""

		self.run = False
		QtWidgets.QMainWindow.show(self)

	def run(self):

		"""
		Does not actual run the experiment, but marks the application for
		running later.
		"""

		self.run = True
		self.options.experiment = str(self.ui.edit_experiment.text())
		self.options.subject = self.ui.spinbox_subject_nr.value()
		self.options.logfile = str(self.ui.edit_logfile.text())
		self.options.fullscreen = self.ui.checkbox_fullscreen.isChecked()
		self.options.custom_resolution = \
			self.ui.checkbox_custom_resolution.isChecked()
		self.options.width = self.ui.spinbox_width.value()
		self.options.height = self.ui.spinbox_height.value()
		self.options.pylink = self.ui.checkbox_pylink.isChecked()
		self.close()


if __name__ == u'__main__':

	from libqtopensesame import __main__
	__main__.opensesamerun()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import re
from libopensesame.oslogging import oslogger
from pyqode.core import api
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'pyqode_manager', category=u'extension')

THRESHOLD = 500


class AutodetectIndentationMode(api.Mode):

	"""Automatically sets indentation based on editor contents."""

	def on_install(self, editor):

		super(AutodetectIndentationMode, self).on_install(editor)
		self.editor.new_text_set.connect(self._autodetect_indentation)
		self._autodetect_indentation()

	def _autodetect_indentation(self):

		if cfg.pyqode_indentation == 'spaces':
			self._set_space_indentation(cfg.pyqode_tab_length)
			return
		if cfg.pyqode_indentation == 'tabs':
			self._set_tab_indentation()
			return
		code = self.editor.toPlainText().replace(u'\u2029', u'\n')
		match_tabs = re.findall(u'((?<=\n)\t+)|(\A\t+)', code)
		tab_length = self._guess_indentation_width(code)
		if tab_length == 0:
			match_spaces = []
		else:
			match_spaces = re.findall(
				u'((?<=\n){indent}+)|(\A{indent}+)'.format(
					indent=u' ' * tab_length
				),
				code
			)
		if len(match_tabs) > len(match_spaces):
			oslogger.debug(u'detected tab-based indentation')
			self._set_tab_indentation()
		else:
			oslogger.debug(
				u'detected space-based indentation (width={})'.format(
					tab_length
				)
			)
			self._set_space_indentation(tab_length)

	def _set_tab_indentation(self):

		self.editor.use_spaces_instead_of_tabs = False

	def _set_space_indentation(self, tab_length):

		self.editor.use_spaces_instead_of_tabs = True
		self.editor.tab_length = tab_length

	def _guess_indentation_width(self, code):
		
		i2 = i3 = i4 = 0
		for line in code.splitlines():
			n_indent = len(line) - len(line.lstrip(u' '))
			if not n_indent % 4:
				i4 += 1
				i3 -= 1
			elif not n_indent % 2:
				i2 += 1
				i4 -= 1
				i3 -= 1
			elif not n_indent % 3:
				i3 += 1
				i2 -= 1
				i4 -= 1
			if i4 - i2 > THRESHOLD and i4 - i3 > THRESHOLD:
				return 4
			if i3 - i2 > THRESHOLD and i3 - i4 > THRESHOLD:
				return 3
			if i2 - i3 > THRESHOLD and i2 - i4 > THRESHOLD:
				return 2
		if i4 > max(i2, i3):
			return 4
		if i3 > max(i2, i4):
			return 3
		if i2 > max(i4, i3):
			return 2
		return cfg.pyqode_tab_length



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.

# About

Provides printing of debug output, which is only shown when OpenSesame is
started in debug mode by passing the `--debug` command-line argument, optionally
with a full stacktrace if the `--stack` argument is passed as well. To print
debug output, simply use the following structure:

	from libopensesame import debug
	debug.print(u'This is a debug message')
"""

from libopensesame.py3compat import *
import sys

stack_lvl = 0
msg_nr = 0

def indent(d=1):

	global stack_lvl
	stack_lvl += d

def _msg(msg=u'', reason=None):

	"""
	desc:
		Prints a debugging message. Respects the --debug and --stack parameters.

	keywords:
		msg:	A debug message.
		reason:	A reason for the message.
	"""

	global stack, max_stack, msg_nr
	if reason is not None:
		msg = u'[%s] %s' % (reason, msg)
	# The terminal may not like anything but plain ASCII
	if not isinstance(msg, basestring):
		msg = str(msg)
	msg = safe_encode(msg, enc=u'ascii', errors=u'ignore')
	try:
		print(u'%.6d %s%s' % (msg_nr, u'| '*stack_lvl, msg))
	except:
		# This should not happen!
		print(u'%.6d %sFailed to print message' % (msg_nr, u'| '*stack_lvl))
	msg_nr += 1

def _print(msg):

	"""
	desc:
		Prints a message to the standard output, just like the normal `print`
		statement/ function. This is necessary to capture encoding errors while
		printing.

	arguments:
		msg:
			desc:	A message to print.
			type:	[unicode, str]
	"""

	try:
		print(msg)
	except:
		if isinstance(msg, bytes):
			print(safe_decode(msg, enc=u'ascii', errors=u'ignore'))
		elif isinstance(msg, str):
			print(safe_encode(msg, enc=u'ascii', errors=u'ignore'))

enabled = '--debug' in sys.argv or '-d' in sys.argv
if enabled:
	msg = _msg
else:
	# Replace the message function with a dummy function to turn off debugging
	# output
	msg = lambda msg=None, reason=None: None



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.mouse_response import (
	mouse_response as mouse_response_runtime
)
from libqtopensesame.items.qtplugin import qtplugin
from libqtopensesame.validators import timeout_validator
from libqtopensesame.misc.translate import translation_context
from libqtopensesame._input.item_combobox import item_combobox
from libqtopensesame.items.feedpad import feedpad
_ = translation_context(u'mouse_response', category=u'item')


class mouse_response(mouse_response_runtime, qtplugin):

	"""mouse_response item GUI"""

	description = _(u'Collects mouse responses')
	help_url = u'manual/response/mouse'
	lazy_init = True

	def __init__(self, name, experiment, string=None):

		"""
		desc:
			Constructor

		arguments:
			name:		The item name.
			experiment: The experiment object.

		keywords:
			string:		The item definition string, or None for default
						initialization.
		"""

		mouse_response_runtime.__init__(self, name, experiment, string)
		qtplugin.__init__(self)

	def init_edit_widget(self):

		"""
		desc:
			Initialize controls.
		"""

		qtplugin.init_edit_widget(self, stretch=True)
		self.add_line_edit_control(
			var=u'correct_response',
			label=_(u'Correct response'),
			info=_(u'Leave empty to use "correct_response"')
		)
		self.add_line_edit_control(
			var=u'allowed_responses',
			label=_(u'Allowed responses'),
			info=_(u'Separated by semicolons, e.g. "left_button;right_button"')
		)
		self.add_line_edit_control(
			var=u'timeout',
			label=_(u'Timeout'),
			tooltip=_(u'In milliseconds or "infinite"'),
			validator=timeout_validator(self)
		)
		self._combobox_sketchpad = item_combobox(
			self.main_window,
			filter_fnc=lambda item: isinstance(
				self.experiment.items[item],
				feedpad
			)
		)
		self._combobox_sketchpad.activated.connect(self.apply_edit_changes)
		self.auto_combobox[u'linked_sketchpad'] = self._combobox_sketchpad
		self.add_control(
			label=_(u'Linked sketchpad'),
			widget=self._combobox_sketchpad,
			info=_('Elements define regions of interest')
		)
		self.add_combobox_control(
			var=u'event_type',
			label=_(u'Event type'),
			options=[
				u'mouseclick',
				u'mouserelease'
			]
		)
		self.add_checkbox_control(
			var=u'show_cursor',
			label=_(u'Visible mouse cursor')
		)
		self.add_checkbox_control(
			var=u'flush',
			label=_(u'Flush pending mouse clicks')
		)

	def edit_widget(self):

		"""See qtplugin."""

		self._combobox_sketchpad.refresh()
		qtplugin.edit_widget(self)

	def rename(self, from_name, to_name):

		"""See qtplugin."""

		qtplugin.rename(self, from_name, to_name)
		if self.var.linked_sketchpad == from_name:
			self.var.linked_sketchpad = to_name

	def delete(self, item_name, item_parent=None, index=None):

		"""See qtplugin."""

		if self.var.linked_sketchpad == item_name:
			self.var.linked_sketchpad = u''



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import debug
from libopensesame.widgets._widget import Widget
from openexp.canvas_elements import RichText


class RatingScale(Widget):

	"""
	desc: |
		The rating_scale widget is a horizontally aligned series of checkable
		boxes (nodes), optionally with a label attached to each node.

		__Example (OpenSesame script):__

		~~~
		widget 0 0 1 1 label text="I like fluffy kittens"
		widget 0 1 1 1 rating_scale var="response" nodes="Agree;Don't know;Disagree"
		~~~

		__Example (OpenSesame script):__

		~~~ .python
		form = Form()
		label = Label(text='I like fluffy kittens')
		rating_scale = RatingScale(nodes=['Agree', "Don't know",
			'Disagree'], var='response')
		form.set_widget(label, (0,0))
		form.set_widget(rating_scale, (0,1))
		form._exec()
		~~~

		[TOC]
	"""

	def __init__(self, form, nodes=5, click_accepts=False,
		orientation=u'horizontal', var=None, default=None):

		"""
		desc: |
			Constructor to create a new `RatingScale` object. You do not
			generally call this constructor directly, but use the
			`RatingScale()` factory function, which is described here:
			[/python/common/]().

		arguments:
			form:
				desc:	The parent form.
				type:	form

		keywords:
			nodes:
				desc:	The number of nodes or a list of node identifiers (e.g.,
						['yes', 'no', 'maybe']. If a list is passed the rating
						scale will have labels, otherwise it will just have
						boxes.
				type:	[int, list]
			click_accepts:
				desc:	Indicates whether the form should close when a value
						is selected.
				type:	bool
			orientation:
				desc:	|
						'horizontal' indicates a horizontally oriented rating
						scale, 'vertical' indicates a vertically oriented rating
						scale.
				type:	[str, unicode]
			var:
				desc:	The name of the experimental variable that should be
						used to log the widget status. The value that is logged
						is the number of the node that was selected, with the
						first node being 0. If no nodes were selected, the value
						is 'None'. For more information about the use of
						response variables in forms, see the form documentation
						page.
				type:	[str, unicode, NoneType]
			default:
				desc:	The node that is selected by default, or `None` to
						select no node. The value corresponds to the node
						number, where 0 is the first node.
				type:	[int, NoneType]
		"""

		if isinstance(click_accepts, basestring):
			click_accepts = click_accepts == u'yes'
		Widget.__init__(self, form)
		self.type = u'rating_scale'
		self.click_accepts = click_accepts
		self.pos_list = []
		self.var = var
		self.orientation = orientation
		if type(nodes) == int:
			self.nodes = [u'']*nodes
		elif isinstance(nodes, basestring):
			self.nodes = nodes.split(u';')
		else:
			self.nodes = nodes
		self.set_value(default)

	def on_mouse_click(self, pos):

		"""
		desc:
			Is called whenever the user clicks on the widget. Selects the
			correct value from the scale and optionally closes the form.

		arguments:
			pos:
				desc:	An (x, y) coordinates tuple.
				type:	tuple
		"""

		for i, (x, y) in enumerate(self.pos_list):
			if not self._inside(pos, (x, y, self.box_size, self.box_size)):
				continue
			self.set_value(i)
			self._update()
			if self.click_accepts:
				return i
			break

	def _init_canvas_elements(self):

		"""
		desc:
			Initializes all canvas elements.
		"""

		x, y, w, h = self.rect
		cx = x+w/2
		cy = y+h/2
		bs = self.form.theme_engine.box_size()
		# The outline
		if self.orientation == 'horizontal':
			box_rect = x, cy-.5*bs, w, 2*bs
		elif self.orientation == 'vertical':
			box_rect = cx-.5*bs, y, 2*bs, h
		else:
			raise osexception((u'rating_scale orientation must be '
				'"horizontal" or "vertical", not "%s"') % self.orientation)
		self.canvas.add_element(
			self.form.theme_engine.frame(*box_rect, style=u'light')
		)
		self._checked_boxes = []
		self._unchecked_boxes = []
		# The distances between nodes
		dx = (1*w-3*bs)/(len(self.nodes)-1)
		dy = (1*h-3*bs)/(len(self.nodes)-1)
		for i, node in enumerate(self.nodes):
			text_width, text_height = RichText(node).construct(self.canvas).size
			if self.orientation == 'horizontal':
				node_x = x + bs + i * dx
				node_y = cy
				text_x = node_x + self.box_size / 2
				text_y = cy - bs - .5 * text_height
				center = True
			elif self.orientation == 'vertical':
				node_x = cx
				node_y = y + bs + i * dy
				text_x = cx + 2 * bs
				text_y = node_y - .5 * text_height + .5 * bs
				center = False
			self.pos_list.append((node_x, node_y))
			self.canvas.add_element(
				RichText(node, center=center, x=text_x, y=text_y)
			)
			cb = self.form.theme_engine.box(node_x, node_y, checked=True)
			ub = self.form.theme_engine.box(node_x, node_y, checked=False)
			self._checked_boxes.append(cb)
			self._unchecked_boxes.append(ub)
			self.canvas.add_element(cb)
			self.canvas.add_element(ub)

	def _update(self):

		"""
		desc:
			Draws the widget.
		"""

		for i, (cb, ub) in enumerate(
			zip(self._checked_boxes, self._unchecked_boxes)):
				cb.visible = i == self.value
				ub.visible = i != self.value

	def set_value(self, val):

		"""
		desc:
			Sets the rating scale value.

		arguments:
			val:
				desc:	The value.
				type:	int
		"""

		if val is not None and (val >= len(self.nodes) or val < 0):
			raise osexception(
				u'Trying to select a non-existing node (%s). Did you specify an incorrect default value?'
				% val)
		self.value = val
		self.set_var(val)


rating_scale = RatingScale



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from openexp._coordinates.psycho import Psycho as PsychoCoordinates
from openexp._mouse.mouse import Mouse
from psychopy import event
from openexp.backend import configurable


class Psycho(Mouse, PsychoCoordinates):

	"""
	desc:
		This is a mouse backend built on top of PsychoPy.
		For function specifications and docstrings, see
		`openexp._mouse.mouse`.
	"""

	def __init__(self, experiment, **resp_args):

		Mouse.__init__(self, experiment, **resp_args)
		PsychoCoordinates.__init__(self)
		self.mouse = event.Mouse(visible=False, win=self.experiment.window)
		event.mouseButtons = [0, 0, 0]

	@configurable
	def get_click(self):

		buttonlist = self.buttonlist
		timeout = self.timeout
		self.mouse.setVisible(self.visible)
		start_time = self.experiment.clock.time()
		button = None
		pos = None
		self.mouse.clickReset()
		while True:
			time = self.experiment.clock.time()
			buttons, times = self.mouse.getPressed(getTime=True)
			for i in (1,2,3):
				if buttons[i-1] and (buttonlist is None or i in buttonlist):
					button = i
					pos = self.mouse.getPos()
					break
			else:
				if timeout is not None and time-start_time >= timeout:
					break
				continue
			break
		if pos is not None:
			pos = self.from_xy(pos)
		self.mouse.setVisible(self._cursor_shown)
		return button, pos, time

	def show_cursor(self, show=True):

		self.mouse.setVisible(show)
		Mouse.show_cursor(self, show=show)

	def get_pos(self):

		return self.from_xy(self.mouse.getPos()), self.experiment.time()

	def set_pos(self, pos=(0,0)):

		self.mouse.setPos(self.to_xy(pos))

	def get_pressed(self):

		return tuple(self.mouse.getPressed(getTime=False))

	def flush(self):

		event.mouseButtons = [0,0,0]
		event.clearEvents()
		return False


# Non PEP-8 alias for backwards compatibility
psycho = Psycho



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame.item import item
from libqtopensesame.items.qtautoplugin import qtautoplugin
import imp

class external_script(item):

	description = \
		u"Run Python code directly from a script file"

	def reset(self):

		"""
		desc:
			Resets plug-in to initial state.
		"""

		self.module = None
		self.var.file = u''
		self.var.prepare_func = u'prepare'
		self.var.run_func = u'run'

	def prepare(self):

		"""
		desc:
			Prepares the item.
		"""

		item.prepare(self)
		if self.module is None:
			path = self.experiment.pool[self.var.file]
			self.experiment.python_workspace.run_file(path)
		self.prepare_bytecode = self.experiment.python_workspace._compile(
			u'%s()' % self.var.prepare_func)
		self.run_bytecode = self.experiment.python_workspace._compile(
			u'%s()' % self.var.run_func)
		self.experiment.python_workspace._exec(self.prepare_bytecode)

	def run(self):

		"""
		desc:
			Runs the item.
		"""

		self.set_item_onset()
		self.experiment.python_workspace._exec(self.run_bytecode)

class qtexternal_script(external_script, qtautoplugin):

	def __init__(self, name, experiment, script=None):

		# Call parent constructors.
		external_script.__init__(self, name, experiment, script)
		qtautoplugin.__init__(self, __file__)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame import debug, misc
from libopensesame.exceptions import osexception
from libqtopensesame.widgets.base_widget import base_widget
from libqtopensesame.misc import theme
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libqtopensesame.runners import RUNNER_LIST, DEFAULT_RUNNER
from qtpy import QtCore, QtWidgets
import os


class preferences_widget(base_widget):

	"""
	desc:
		The widget for the preferences tab.
	"""

	def __init__(self, main_window):

		"""
		desc:
			Constructor.

		arguments:
			main_window:	A qtopensesame object.
		"""

		super(preferences_widget, self).__init__(
			main_window,
			ui=u'widgets.preferences_widget'
		)
		self.tab_name = u'__preferences__'
		self.lock = False
		# Connect the controls
		self.ui.checkbox_toolbar_text.toggled.connect(self.apply)
		self.ui.checkbox_small_toolbar.toggled.connect(self.apply)
		self.ui.combobox_runner.currentIndexChanged.connect(self.apply)
		self.ui.combobox_style.currentIndexChanged.connect(self.apply)
		self.ui.combobox_theme.currentIndexChanged.connect(self.apply)
		self.ui.combobox_locale.currentIndexChanged.connect(self.apply)
		self.set_controls()
		for ext in self.extensions:
			try:
				w = ext.settings_widget()
			except Exception as e:
				if not isinstance(e, osexception):
					e = osexception(msg=u'Extension error', exception=e)
				self.notify(
					(
						u'Extension %s failed to return settings widget '
						u'(see debug window for stack trace)'
					) % ext.name()
				)
				self.console.write(e)
				continue
			if w is None:
				continue
			if hasattr(w, u'__advanced__'):
				self.ui.layout_advanced.addRow(w, None)
			else:
				self.ui.layout_preferences.addRow(w, None)
		self.ui.container_widget.adjustSize()

	def set_controls(self):

		"""
		desc:
			Updates the controls.
		"""

		if self.lock:
			return
		self.lock = True
		self.ui.checkbox_toolbar_text.setChecked(
			self.main_window.ui.toolbar_main.toolButtonStyle() ==
			QtCore.Qt.ToolButtonTextUnderIcon
		)
		self.ui.checkbox_small_toolbar.setChecked(cfg.toolbar_size == 16)
		try:
			runner_index = RUNNER_LIST.index(cfg.runner)
		except ValueError:
			oslogger.warning('invalid runner: {}'.format(cfg.runner))
			runner_index = RUNNER_LIST.index(DEFAULT_RUNNER)
		self.ui.combobox_runner.setCurrentIndex(runner_index)
		# Set the locale combobox
		self.ui.combobox_locale.addItem(u'[Default]')
		self.ui.combobox_locale.setCurrentIndex(0)
		locales = sorted(
			[
				locale[:-3]
				for locale in os.listdir(misc.resource(u'locale'))
				if locale != u'translatables.qm'
			] + [u'en_US']
		)
		for i, locale in enumerate(locales):
			self.ui.combobox_locale.addItem(locale)
			if cfg.locale == locale:
				self.ui.combobox_locale.setCurrentIndex(i + 1)
		# Set the style combobox
		self.ui.combobox_style.addItem(u"[Default]")
		self.ui.combobox_style.setCurrentIndex(0)
		for i, style in enumerate(QtWidgets.QStyleFactory.keys()):
			self.ui.combobox_style.addItem(style)
			if cfg.style == style:
				self.ui.combobox_style.setCurrentIndex(i + 1)
		# Set the theme combobox
		for i, _theme in enumerate(theme.available_themes):
			self.ui.combobox_theme.addItem(_theme)
			if cfg.theme == _theme:
				self.ui.combobox_theme.setCurrentIndex(i)
		self.lock = False

	def apply(self):

		"""Apply the controls"""

		if self.lock:
			return
		self.lock = True
		self.main_window.ui.toolbar_main.setToolButtonStyle(
			QtCore.Qt.ToolButtonTextUnderIcon
			if self.ui.checkbox_toolbar_text.isChecked() else
			QtCore.Qt.ToolButtonIconOnly
		)
		# Apply locale
		cfg.locale = self.ui.combobox_locale.currentText()
		if cfg.locale == u'[Default]':
			cfg.locale = u''
		# Apply toolbar size
		old_size = cfg.toolbar_size
		new_size = 16 if self.ui.checkbox_small_toolbar.isChecked() else 32
		if old_size != new_size:
			cfg.toolbar_size = new_size
			self.theme.set_toolbar_size(cfg.toolbar_size)
		# Apply runner
		cfg.runner = RUNNER_LIST[self.ui.combobox_runner.currentIndex()]
		# Apply theme and style
		cfg.theme = self.ui.combobox_theme.currentText()
		cfg.style = self.ui.combobox_style.currentText()
		self.main_window.save_state()
		self.lock = False



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'toolbar_menu', category=u'extension')


class toolbar_menu(base_extension):

	"""
	desc:
		Integrates the menu into the toolbar.
	"""

	def event_startup(self):

		self._menu = None
		if not cfg.toolbar_menu_active:
			return
		cfg.toolbar_menu_active = False  # It will be toggled!
		self.activate()
		self.set_checked(True)

	def activate(self):

		"""
		desc:
			Toggle the menubar integration.
		"""

		if self._menu is None:
			self._init_menu()
		if cfg.toolbar_menu_active:
			self._deactivate_toolbar_menu()
		else:
			self._activate_toolbar_menu()
			
	def _init_menu(self):
		
		"""
		desc:
			Creates a widget with a menu that copies all actions from the main
			menu bar.
		"""		
			
		self._menu = QtWidgets.QMenu()
		for action in self.menubar.actions():
			self._menu.addAction(action)
		self.stretch = QtWidgets.QWidget()
		self.stretch.setSizePolicy(
			QtWidgets.QSizePolicy.Expanding,
			QtWidgets.QSizePolicy.Expanding
		)
		self.button = QtWidgets.QPushButton(
			self.theme.qicon(self.icon()),
			_(u'Menu')
		)
		self.button.setMenu(self._menu)
		self.button.setIconSize(
			QtCore.QSize(cfg.toolbar_size, cfg.toolbar_size)
		)
		self.button.setFlat(True)
		self.toolbar.addWidget(self.stretch)
		self.menu_action = self.toolbar.addWidget(self.button)
			
	def _keep_toolbar_visible(self, visible):
		
		if visible:
			return
		oslogger.debug('keep toolbar visible')
		self.toolbar.setVisible(True)

	def _activate_toolbar_menu(self):

		"""
		desc:
			Hide the menu bar and show the toolbar widget.
		"""

		cfg.toolbar_menu_active = True
		self.menubar.setVisible(False)
		self.menu_action.setVisible(True)
		self.stretch.setVisible(True)
		if not self.toolbar.isVisible():
			self.toolbar.setVisible(True)
		self.toolbar.visibilityChanged.connect(self._keep_toolbar_visible)

	def _deactivate_toolbar_menu(self):

		"""
		desc:
			Show the menu bar and hide the toolbar widget.
		"""

		cfg.toolbar_menu_active = False
		self.menubar.setVisible(True)
		self.menu_action.setVisible(False)
		self.stretch.setVisible(False)
		self.toolbar.visibilityChanged.disconnect()



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import item
from datamatrix import operations, DataMatrix, functional
from pseudorandom import Enforce, MaxRep, MinDist
import openexp.keyboard


class loop(item.item):

	"""A loop item runs a single other item multiple times"""

	description = u'Repeatedly runs another item'
	valid_orders = u'sequential', u'random'
	commands = [
		u'fullfactorial',
		u'shuffle',
		u'shuffle_horiz',
		u'slice',
		u'sort',
		u'sortby',
		u'reverse',
		u'roll',
		u'weight',
	]

	def reset(self):

		"""See item."""

		self.ef = None
		self.dm = DataMatrix(length=0)
		self.dm.sorted = False
		self.live_dm = None
		self.live_row = None
		self._operations = []
		self._constraints = []
		self._item = u''

		self.var.repeat = 1
		self.var.continuous = u'no'
		self.var.order = u'random'
		self.var.break_if = u'never'
		self.var.break_if_on_first = u'yes'
		self.var.source = u'table'  # file or table
		self.var.source_file = u''

	def from_string(self, string):

		"""See item."""

		self.var.clear()
		self.comments = []
		self.reset()
		if string is None:
			return
		for i in string.split(u'\n'):
			self.parse_variable(i)
			cmd, arglist, kwdict = self.syntax.parse_cmd(i)
			if cmd == u'run':
				if len(arglist) != 1 or kwdict:
					raise osexception(u'Invalid run command: %s' % i)
				self._item = arglist[0]
				continue
			if cmd == u'setcycle':
				if self.ef is not None or self._operations:
					raise osexception(
						u'setcycle must come before constraints and operations')
				if len(arglist) != 3 or kwdict:
					raise osexception(u'Invalid setcycle command: %s' % i)
				row, var, val = tuple(arglist)
				if row >= len(self.dm):
					self.dm.length = row + 1
				if var not in self.dm:
					self.dm[var] = u''
				self.dm[row][var] = val
				continue
			if cmd == u'constrain':
				if self._operations:
					raise osexception(
						u'constraints must come before operations'
					)
				if len(arglist) != 1:
					raise osexception(u'Invalid constrain command: %s' % i)
				colname = arglist[0]
				for constraint, value in kwdict.items():
					if constraint == u'maxrep':
						constraint_cls = MaxRep
						kwargs = {u'maxrep': value}
					elif constraint == u'mindist':
						constraint_cls = MinDist
						kwargs = {u'mindist': value}
					else:
						raise osexception(
							u'Unknown constraint: %s' % constraint
						)
					self._constraints.append((
						constraint_cls,
						colname,
						kwargs
					))
					# raise osexception(u'Invalid constrain command: %s' % i)
				continue
			if cmd in self.commands:
				self._operations.append((cmd, arglist))
		if len(self.dm) == 0:
			self.dm.length = 1
		if len(self.dm.columns) == 0:
			self.dm.empty_column = u''
		# Backwards compatibility: Older version of OpenSesame can specify the
		# number of cycles through the cycles variable. If the specified number
		# of cycles doesn't match the length of the datamatrix, we change the
		# length of the datamatrix.
		if u'cycles' in self.var and isinstance(self.var.cycles, int) \
			and self.var.cycles != len(self.dm):
			self.dm.length = self.var.cycles

	def to_string(self):

		"""See item."""

		# Older versions of OpenSesame use an explicit cycles variable, so we
		# set this here for backwards compatibility.
		self.var.cycles = len(self.dm)
		s = item.item.to_string(self)
		for i, row in enumerate(self.dm):
			for name, val in row:
				s += u'\t%s\n' % \
					self.syntax.create_cmd(u'setcycle', [i, name, val])
		for constraint_cls, colname, kwargs in self._constraints:
			s += u'\t%s\n' % self.syntax.create_cmd(
				u'constrain', [colname], kwargs
			)
		for cmd, arglist in self._operations:
			s += u'\t%s\n' % self.syntax.create_cmd(cmd, arglist)
		s += u'\t%s\n' % self.syntax.create_cmd(u'run', [self._item])
		return s

	def _require_arglist(self, cmd, arglist, minlen=1):

		"""
		desc:
			Checks whether a non-empty argument list has been specified, and
			raises an Exception otherwise.

		arguments:
			cmd:
				desc:	The command to check for.
				type:	str
			arglist:
				desc:	The argument list to check.
				type:	list
		"""

		if len(arglist) < minlen:
			raise osexception(u'Invalid argument list for %s' % cmd)

	def _create_live_datamatrix(self):

		"""
		desc:
			Builds a live DataMatrix. That is, it takes the orignal DataMatrix
			and applies all the operations as specified.

		returns:
			desc:	A live DataMatrix.
			type:	DataMatrix
		"""

		src_dm = self.dm if self.var.source == u'table' else self._read_file()
		for column_name in src_dm.column_names:
			if not self.syntax.valid_var_name(column_name):
				raise osexception(
					u'The loop table contains an invalid column name: 'u'\'%s\''
					% column_name
				)
		# The number of repeats should be numeric. If not, then give an error.
		# This can also occur when generating a preview of a loop table if
		# repeat is variable.
		if not isinstance(self.var.repeat, (int, float)):
			raise osexception(
				u'Don\'t know how to generate a DataMatrix for "%s" repeats'
				% self.var.repeat
			)
		length = int(len(src_dm) * self.var.repeat)
		dm = DataMatrix(length=0)
		while len(dm) < length:
			i = min(length-len(dm), len(src_dm))
			if self.var.order == u'random':
				dm <<= operations.shuffle(src_dm)[:i]
			else:
				dm <<= src_dm[:i]
		if self.var.order == u'random':
			dm = operations.shuffle(dm)
		# Constraints come before loop operations
		if self._constraints:
			self.ef = Enforce(dm)
			for constraint_cls, colname, kwargs in self._constraints:
				colname = self.syntax.auto_type(self.syntax.eval_text(colname))
				kwargs = {
					key: self.syntax.auto_type(self.syntax.eval_text(val))
					for key, val in kwargs.items()
				}
				try:
					cols = dm[colname]
				except AttributeError:
					raise osexception(
						u'Column %s does not exist' % colname
					)
				self.ef.add_constraint(
					constraint_cls, cols=cols, **kwargs
				)
			dm = self.ef.enforce()
		# Operations come last
		for cmd, arglist in self._operations:
			arglist = [
				self.syntax.auto_type(self.syntax.eval_text(arg))
				for arg in arglist
			]
			# The column name is always specified last, or not at all
			if arglist:
				try:
					colname = arglist[-1]
					if not isinstance(colname, int):
						col = dm[colname]
				except (IndexError, AttributeError):
					raise osexception(
						u'Column %s does not exist' % arglist[-1]
					)
			if cmd == u'fullfactorial':
				try:
					dm = operations.fullfactorial(dm)
				except MemoryError:
					raise osexception(u'DataMatrix too large for fullfact')
			elif cmd == u'shuffle':
				if not arglist:
					dm = operations.shuffle(dm)
				else:
					dm[colname] = operations.shuffle(col)
			elif cmd == u'shuffle_horiz':
				if not arglist:
					dm = operations.shuffle_horiz(dm)
				else:
					# There can be multiple column names, so we need to check
					# if all of them exist, rather than only the last one as
					# we did above.
					for _colname in arglist:
						try:
							dm[_colname]
						except AttributeError:
							raise osexception(
								u'Column %s does not exist'
								% _colname
							)
					dm = operations.shuffle_horiz(
						*[dm[_colname] for _colname in arglist]
					)
			elif cmd == u'slice':
				self._require_arglist(cmd, arglist, minlen=2)
				dm = dm[arglist[0]: arglist[1]]
			elif cmd == u'sort':
				self._require_arglist(cmd, arglist)
				dm[colname] = operations.sort(col)
			elif cmd == u'sortby':
				self._require_arglist(cmd, arglist)
				dm = operations.sort(dm, by=col)
			elif cmd == u'reverse':
				if not arglist:
					dm = dm[::-1]
				else:
					dm[colname] = col[::-1]
			elif cmd == u'roll':
				self._require_arglist(cmd, arglist)
				steps = arglist[0]
				if not isinstance(steps, int):
					raise osexception(u'roll steps should be numeric')
				if len(arglist) == 1:
					dm = dm[-steps:] << dm[:-steps]
				else:
					dm[colname] = list(col[-steps:]) + list(col[:-steps])
			elif cmd == u'weight':
				self._require_arglist(cmd, arglist)
				# Evaluate the weights before passing them to the weight 
				# function, so that weights can be defined in terms of 
				# variables.
				ecol = functional.map_(
					lambda w: self.syntax.auto_type(self.syntax.eval_text(w)),
					col
				)
				try:
					dm = operations.weight(ecol)
				except TypeError:
					raise osexception(
						u'weight values should be non-negative numeric values'
					)
		return dm

	def prepare(self):

		"""See item."""

		item.item.prepare(self)
		# Deprecation errors. The direct reference to __vars__ prevents a lookup
		# in the parent var store.
		if (u'skip' in self.var.__vars__ and self.var.skip != 0) \
			or (u'offset' in self.var.__vars__ and self.var.offset != u'no'):
			raise osexception(
				u'The skip and offset options have been removed. Please refer '
				u'to the documentation of the loop item for more information.'
			)
		# Compile break-if statement
		break_if = self.var.get(u'break_if', _eval=False)
		if break_if not in (u'never', u''):
			try:
				self._break_if = self.syntax.compile_cond(break_if)
			except osexception as e:
				raise osexception(
					u'Error compiling break-if expression',
					item=self.name,
					exception=e
				)
		else:
			self._break_if = None
		# Create a keyboard to flush responses between cycles
		self._keyboard = openexp.keyboard.keyboard(self.experiment)
		# Make sure the item to run exists
		if self._item not in self.experiment.items:
			raise osexception(
				u"Could not find item '%s', which is called by loop item '%s'"
				% (self._item, self.name)
			)

	def run(self):

		"""See item."""

		self.set_item_onset()
		if self.live_dm is None or self.var.continuous == u'no':
			self.live_dm = self._create_live_datamatrix()
			self.live_row = 0
		first = True
		while self.live_row < len(self.live_dm):
			self.experiment.var.repeat_cycle = 0
			self.experiment.var.live_row = self.live_row
			self.experiment.var.set('live_row_%s' % self.name, self.live_row)
			for name, val in self.live_dm[self.live_row]:
				if isinstance(val, basestring) and val.startswith(u'='):
					try:
						val = self.python_workspace._eval(val[1:])
					except Exception as e:
						raise osexception(
							u'Error evaluating Python expression in loop table',
							line_offset=0,
							item=self.name,
							phase=u'run',
							exception=e
						)
				self.experiment.var.set(name, val)
			# Evaluate the run if statement
			if (
				self._break_if is not None and
				(not first or self.var.break_if_on_first == u'yes')
			):
				self.python_workspace[u'self'] = self
				try:
					if self.python_workspace._eval(self._break_if):
						break
				except osexception as e:
					raise osexception(
						u'Error evaluating break-if expression',
						item=self.name,
						exception=e
					)
			# Run the item!
			self.experiment.items.execute(self._item)
			# If the repeat_cycle flag was set, run the item again later
			if self.experiment.var.repeat_cycle:
				self.live_dm <<= self.live_dm[self.live_row:self.live_row+1]
				if self.var.order == u'random':
					self.live_dm = self.live_dm[:self.live_row+1] \
						<< operations.shuffle(self.live_dm[self.live_row+1:])
			self.live_row += 1
			first = False
		else:
			# If the loop finished without breaking, it needs to be reset on
			# the next run of the loop item
			self.live_row = None
			self.live_dm = None

	def _read_file(self):

		"""
		desc:
			Reads a source file and raises an osexception if this fails.

		returns:
			type:	DataMatrix
		"""

		from datamatrix import io
		src = self.experiment.pool[self.var.source_file]
		if src.endswith(u'.xlsx'):
			try:
				return io.readxlsx(src)
			except Exception as e:
				raise osexception(u'Failed to read .xlsx file: %s' % src,
					exception=e)
		try:
			return io.readtxt(src)
		except Exception as e:
			raise osexception(
				(u'Failed to read text file (perhaps it has the '
				u'wrong format or it is not utf-8 encoded): %s') % src,
				exception=e
			)

	def _var_info_table(self):

		"""
		returns:
			A list of (var, value) tuples that have been defined in the loop
			table.
		"""

		return [(colname, safe_decode(col)) for colname, col in self.dm.columns]

	def _var_info_file(self):

		"""
		returns:
			A list of (var, value) tuples that have been defined in a source
			file.
		"""

		try:
			dm = self._read_file()
		except osexception:
			return []
		return [(colname, safe_decode(col)) for colname, col in dm.columns]

	def var_info(self):

		"""See item."""

		return item.item.var_info(self) + (
			self._var_info_table()
			if self.var.source == u'table'
			else self._var_info_file()
		)



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.extensions import base_extension
from libqtopensesame.misc.config import cfg
from libopensesame import metadata
from qtpy import QtCore
import sys
import yaml
import os
import traceback
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'bug_report', category=u'extension')

class bug_report(base_extension):

	"""
	desc:
		Captures the stderr and submits bug reports based on it.
	"""

	def event_startup(self):

		"""
		desc:
			Start capturing.
		"""
		sys.excepthook = self.captured_err

	def indent(self, s):

		"""
		desc:
			Tab-indent a piece of text so that it's code for Markdown.

		arguments:
			s:	The text to indent.

		returns:
			Indented text.
		"""

		return u'\t' + s.replace(u'\n', u'\n\t').replace(os.linesep, u'\n\t')

	def event_bug_report_send(self):

		"""
		desc:
			Sends a bug report for the latest stacktrace. Also closes the
			current tab, which is the report tab, and shows a results tab.
		"""

		self.main_window.tabwidget.close_current()
		if py3:
			from urllib.request import urlopen
			from urllib.parse import urlencode
		else:
			from urllib import urlencode
			from urllib2 import urlopen
		if self.traceback is None:
			return
		q = urlencode({
			u'traceback' : safe_str(self.traceback, errors=u'ignore'),
			u'version' : safe_str(metadata.__version__, errors=u'ignore'),
			u'python_version' : safe_str(metadata.python_version,
				errors=u'ignore'),
			u'platform' : safe_str(metadata.platform, errors=u'ignore'),
			})
		url = cfg.bug_report_url + u'?' + q
		try:
			fd = urlopen(url)
			resp = safe_decode(fd.read(), errors=u'ignore')
			fd.close()
		except:
			self.tabwidget.open_markdown(self.ext_resource(u'failure.md'),
				title=_(u'Bug report not sent'))
			return
		if resp == u'OK':
			self.tabwidget.open_markdown(self.ext_resource(u'success.md'),
				title=_(u'Bug report sent'))
		else:
			self.tabwidget.open_markdown(self.ext_resource(u'failure.md'),
				title=_(u'Bug report not sent'))

	def captured_err(self, exception_type, value, tb):

		"""
		desc:
			Shows a report tab when an error message has been captured.
		"""

		error_list = traceback.format_exception(exception_type, value, tb)
		self.traceback = u"".join([safe_decode(tb_line) for tb_line in error_list])
		self.traceback_md = u'~~~ .traceback\n%s\n~~~\n' % self.traceback
		md = safe_read(self.ext_resource(u'report.md')) % {
			# u'traceback' : self.indent(self.stderr.buffer),
			u'traceback' : self.traceback_md,
			u'version' : metadata.__version__,
			u'python_version' : safe_str(metadata.python_version,
				errors=u'ignore'),
			u'platform' : metadata.platform,
			}
		self.tabwidget.open_markdown(md, title=_(u'Oops ...'))
		sys.stderr.write(self.traceback)
